VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "LightEngine8"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
'##CD The LightEngine class allows you to add, modify, delete dynamic lights. Dynamic lights are calculated by Direct3D. They're fast _
but sometimes they are less accurate than the lightmaps. If you want lightmaps, create simply your dynamic lights (only PointLight please) and call the GenerateLightmap in mesh or landscape class.

Private MatFac As New MaterialFactory8

Function CreateMaterial() As Integer
   '##BD Register a new material and returns the material index. You have to store the index to change the material settings later.
    CreateMaterial = MatFac.CreateMaterial
End Function

Public Function CreatePointLight(Position As D3DVECTOR, Radius As Single, r, g, b, a) As Integer
  '##BD Creates a directional point light. This the simpler light type . This function returns the light index, you have to store it if your want to change the light later. _
  <i><b>Note :</b> Use this kind of light to create lightmaps</i>
  '##PD Position The position of the light.
  '##PD Radius Range (or radius) of the light
  '##PD Red value of the light (0-1)
  '##PD Green value of the light (0-1)
  '##PD Blue value of the light (0-1)
  '##PD Alpha value of the light (0-1)
   'On Error Resume Next
   Dim i As Integer
   For i = 1 To 100
    If Light(i).Active = False Then Exit For
   Next i
   CreatePointLight = i
   Index = i
   Light(i).Active = True
   With Light(i).Light
    .Ambient = DXColor(r, g, b, a)
    .Diffuse = DXColor(r, g, b, a)
    .range = Radius
    .Attenuation0 = 1
    .Attenuation1 = 0
    .Attenuation2 = 0
    .Position = Vector(0, 0, 0)
    .Type = D3DLIGHT_POINT
     
   'DebugLight i
   Device.SetLight Index, Light(i).Light
   'DebugLight i
   Device.LightEnable Index, True
   .Position = Position
   Device.SetLight Index, Light(i).Light
   

   End With
   
End Function
Public Function CreateDirectionalLight(Orientation As D3DVECTOR, r, g, b, a) As Integer
  '##BD Creates a directional dynamic light. The sun is a directional light. This function returns the light index, you have to store it if your want to change the light later. _
  <i><b>Note :</b> The lightmaps can't be created from directional light, use rather point light</i>
  '##PD Orientation The direction of the light.
  '##PD Red value of the light (0-1)
  '##PD Green value of the light (0-1)
  '##PD Blue value of the light (0-1)
  '##PD Alpha value of the light (0-1)
  
   For i = 1 To 100
    If Light(i).Active = False Then Exit For
   Next i
   CreateDirectionalLight = i
   Index = i
   With Light(i).Light
    .Ambient = DXColor(r, g, b, a)
    .Diffuse = DXColor(r, g, b, a)
    '.range = Radius
    .Attenuation0 = 1
    .Attenuation1 = 1
    .Attenuation2 = 0
    .Type = D3DLIGHT_DIRECTIONAL
    .Direction = Orientation
   End With
   Device.SetLight Index, Light(i).Light
   Device.LightEnable Index, True
End Function

Private Sub DebugLight(id As Integer)
Dim li As D3DLIGHT8

'Device.GetLight Id, li
With li
   Debug.Print .Ambient.r, .Ambient.g, .Ambient.b, .Ambient.a
   Debug.Print .Diffuse.r, .Diffuse.g, .Diffuse.b, .Diffuse.a
   Debug.Print .Position.x, .Position.y, .Position.z
   Debug.Print .range, .Attenuation0, .Attenuation1, .Attenuation2
End With
End Sub

Public Function DeleteLight(LightIndex As Integer)
  '##BD This method deletes a light and frees the light index
  '##PD LightIndex Index of the light to delete.
  Light(LightIndex).Active = False
  Light(LightIndex).Enabled = False
  
  'EnableLight LightIndex, 0
  Dim li As D3DLIGHT8
  Device.SetLight LightIndex, li
  Device.LightEnable LightIndex, 0
  
End Function


Function DeleteMaterial(MaterialIndex As Integer)
  '##BD This method deletes the specified material and frees the material index
  '##PD MaterialIndex Index of the material to delete.
  MatFac.DeleteMaterial MaterialIndex
  
End Function

Public Sub EnableLight(LightIndex As Integer, Enabled As Boolean)
 '##BD Enables/Disables the specified light.
 '##PD LightIndex Index of the light to enable/disable.
 '##PD enable/disable.
 Device.LightEnable LightIndex, Enabled
End Sub




Sub SetLightAmbientColor(Index, r, g, b, a)
 '##BD Set the ambient color of the specified light. Use rather SetLightColor to change a light color.
 '##PD Index Light index
 '##PD r Red value (0-1)
 '##PD g Green value (0-1)
 '##PD b Blue value (0-1)
 '##PD a Alpha value (0-1)
 Light(Index).Light.Ambient = DXColor(r, g, b, a)
 Device.SetLight Index, Light(Index).Light
End Sub
Sub SetLightDiffuseColor(Index, r, g, b, a)
 '##BD Change the diffuse color of a specified light. Use rather SetLightColor to change a light color.
  '##PD Index Light index
 '##PD r Red value (0-1)
 '##PD g Green value (0-1)
 '##PD b Blue value (0-1)
 '##PD a Alpha value (0-1)
 Light(Index).Light.Diffuse = DXColor(r, g, b, a)
 Device.SetLight Index, Light(Index).Light
End Sub
Sub SetLightDirection(Index, x As Single, y As Single, z As Single)
 '##BD Changes the direction of a directional light
 '##PD x New direction x-coordinate
 '##PD y New direction y-coordinate
 '##PD z New direction z-coordinate
 
 Light(Index).Light.Direction = Vector(x, y, z)
 Device.SetLight Index, Light(Index).Light
End Sub

Private Sub SetLightSpecularColor(Index, r, g, b, a)
  '##BD Change the specular color of a specified light. Use rather SetLightColor to change a light color.
  '##PD Index Light index
 '##PD r Red value (0-1)
 '##PD g Green value (0-1)
 '##PD b Blue value (0-1)
 '##PD a Alpha value (0-1)
  Light(Index).Light.Specular = DXColor(r, g, b, a)
 Device.SetLight Index, Light(Index).Light
End Sub

Sub SetLightFalloff(Index, Falloff)
  '##BD Not used.
  Light(Index).Light.Falloff = Falloff
 Device.SetLight Index, Light(Index).Light
 
End Sub
Sub SetLightColor(Index, r, g, b, a)
  '##BD Change the color of a light
  '##PD Index Light index (returned by CreateLight's functions)
    '##PD Index Light index
 '##PD r Red value (0-1)
 '##PD g Green value (0-1)
 '##PD b Blue value (0-1)
 '##PD a Alpha value (0-1)
  Light(Index).Light.Ambient = DXColor(r, g, b, a)
    Light(Index).Light.Diffuse = DXColor(r, g, b, a)
 Device.SetLight Index, Light(Index).Light
End Sub
Sub SetLightAttenuation(Index, a1, a2, a3)
   '##BD Set the different attenuation of the specified light.
   '##PD Index Index of the light to modify
   '##PD a1 Attenuation 1
   '##PD a2 Attenuation 2
   '##PD a3 Attenuation 3
   Light(Index).Light.Attenuation0 = a1
   Light(Index).Light.Attenuation1 = a2
   Light(Index).Light.Attenuation2 = a3
   Device.SetLight Index, Light(Index).Light
End Sub
Sub SetLightPosition(LightIndex As Integer, Position As D3DVECTOR)
  '##BD Changes the position of a point or spot light.
  '##PD x New position x-coordinate
  '##PD y New position y-coordinate
  '##PD z New position z-coordinate

  Light(LightIndex).Light.Position.x = Position.x
  Light(LightIndex).Light.Position.y = Position.y
  Light(LightIndex).Light.Position.z = Position.z
  
  'DebugLight LightIndex
  'Dim li As D3DLIGHT8
  'Device.SetLight LightIndex, li
  Device.SetLight LightIndex, Light(LightIndex).Light
  'DebugLight LightIndex
 
End Sub

Sub SetLightRadius(Index As Integer, Radius As Single)
  '##BD Changes the radius (or range) of a point light.
  '##PD Index Light index (returned by CreateLight's function)
  '##PD Radius Range or radius.
  Light(Index).Light.range = Radius
  Device.SetLight Index, Light(Index).Light
End Sub
Function SetMaterialAmbient(Index, r, g, b, a)
  '##BD Change or set the ambient color of the specified material. Ambient color is the color that will be used without lights.
  '##PD Index Material index (returned by CreateMaterial)
  '##PD R Red value (0-1)
  '##PD G Green value (0-1)
  '##PD B Blue value (0-1)
  '##PD A Alpha value (0-1)
  Materials(Index).Mat.Ambient = DXColor(r, g, b, a)
End Function
Function SetMaterialDiffuse(Index, r, g, b, a)
  '##BD Change or set the diffuse color of the specified material. Diffuse color is the real color that will be diffused with lights.
  '##PD Index Material index (returned by CreateMaterial)
  '##PD R Red value (0-1)
  '##PD G Green value (0-1)
  '##PD B Blue value (0-1)
  '##PD A Alpha value (0-1)
  Materials(Index).Mat.Diffuse = DXColor(r, g, b, a)
End Function
Function SetMaterialEmissive(Index, r, g, b, a)
  '##BD Change or set the emissive color of the specified material. Emissive color is the color that will be added with lights.
  '##PD Index Material index (returned by CreateMaterial)
  '##PD R Red value (0-1)
  '##PD G Green value (0-1)
  '##PD B Blue value (0-1)
  '##PD A Alpha value (0-1)
  Materials(Index).Mat.emissive = DXColor(r, g, b, a)
End Function
Function SetMaterialSpecular(Index, r, g, b, a)
   '##BD Change or set the specular color of the specified material. Specular color allows you to create chrome or metal effect.
  '##PD Index Material index (returned by CreateMaterial)
  '##PD R Red value (0-1)
  '##PD G Green value (0-1)
  '##PD B Blue value (0-1)
  '##PD A Alpha value (0-1)
  Materials(Index).Mat.Specular = DXColor(r, g, b, a)
End Function
Function SetMaterialPower(Index, power)
  '##BD Change or set the power of the specified material. This setting allows you to do metal or chrome effect.
  '##PD Index Material index (returned by CreateMaterial)
  '##PD R Red value (0-1)
  '##PD G Green value (0-1)
  '##PD B Blue value (0-1)
  '##PD A Alpha value (0-1)
  Materials(Index).Mat.power = power
End Function

Private Sub Class_Terminate()
  For i = 0 To NBLight
  Next i
  
End Sub


