VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "Mesh8"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
'----------------------
' TrueVisionMesh8 Class
'----------------------
Option Explicit

Private MeshIndex As Integer

Private Type tdsHeader
    Header As Integer
    Length As Long
End Type

Private Type aFace
    a As Integer
    b As Integer
    C As Integer
    Flags As Integer
End Type

Private Type UVmap
    U As Single
    V As Single
End Type

Private Type udColorByte
    Red As Byte
    Green As Byte
    Blue As Byte
End Type

Private Type aMaterial
    MatName As String
    Ambient As D3DCOLORVALUE
    Diffuse As D3DCOLORVALUE
    Specular As D3DCOLORVALUE
    MapName As String       ' Diffuse Map
    MapPercent As Integer   ' Percentage map is used
    MaterialIndex As Integer
End Type
Private Enum colorBlocks
    Ambient
    Diffuse
    Specular
End Enum

Private GroupPointer As Integer

Public Enum CONST_TV_PRIMITIVETYPE
  '##EM Primitive type of the mesh. You can have triangle strips or triangle list.
  TV_TRIANGLESTRIP = 5
  TV_TRIANGLELIST = 4
End Enum

Public Enum CONST_TV_VERTEXTYPE
  TV_COLOR_VERTEX = TV_LVERTEX2
  TV_NORMAL_VERTEX = TV_VERTEX2
  TV_NON_LIT_VERTEX = TV_VERTEX2
End Enum
Function DuplicateMesh(Optional NewMeshName As String = "") As Mesh8
   Dim id As Integer
   
   Set DuplicateMesh = LocalScene.CreateMeshBuilder(NewMeshName)
   
   id = DuplicateMesh.GetMeshIndex
   
   Mesh(id) = Mesh(MeshIndex)
   
End Function


Sub LookAtPoint(Point As D3DVECTOR)
    '##BD Rotates the mesh so that it looks at the specified point.
    '##PD Point The vector that the mesh must look at.
    Dim a As Single
    Dim V As D3DVECTOR
    V = VSubtract(Point, Mesh(MeshIndex).Features.Settings.Position)
      If V.z > 0 Then
        a = Atn(V.x / V.z)
      ElseIf V.z = 0 Then
        a = 0
      Else
        a = Atn(V.x / V.z) + 3.141592
      End If
    Mesh(MeshIndex).Features.Settings.Rotation.y = a * 180 / 3.141592
      Mesh(MeshIndex).Changed = True
End Sub

Sub MoveForward(StepPerSecond As Single, Optional FollowLandscape As Boolean, Optional AltitudeOverLandscape As Single = 5, Optional CheckCollision As Boolean)
 '##BD Makes the mesh move forward with the current rotation settings.
 '##PD StepPerSecond Distance that the mesh have to move in a second.
 '##PD FollowLandscape Decides if the mesh have to follow the landscape
 '##PD AltitudeOverLandscape Define the altitude at which the mesh must be when it follows the landscape.
 Dim NewPos As D3DVECTOR
 Dim step As Single
 step = StepPerSecond * SystemFrame / 1000
 NewPos = Vector(GetPosition.x + Cos(GetRotation.y * PiFactor) * step, GetPosition.y, Cos(GetRotation.y * PiFactor) * step + GetPosition.z)
 
 If CheckCollision = True Then
   If LocalScene.Collide(Mesh(MeshIndex).Features.Settings.Position, NewPos) = True Then
     Exit Sub
   End If
 End If
 SetPosition GetPosition.x + Cos(GetRotation.y * PiFactor) * step, GetPosition.y, Cos(GetRotation.y * PiFactor) * step + GetPosition.z
 If FollowLandscape = True Then
   Mesh(MeshIndex).Features.Settings.Position.y = Land.GetHeight(GetPosition.x, GetPosition.z) + AltitudeOverLandscape
 End If
   Mesh(MeshIndex).Changed = True
End Sub

Function Collision(Origin As D3DVECTOR, EndPos As D3DVECTOR, Optional CheckType As CONST_TV_TESTTYPE = CONST_TV_TESTTYPE.TV_TESTTYPE_ACCURATETESTING) As Boolean
 '##BD Check if the line (Origin To Endpos) intersects the mesh. You can choose the collision type, bounding box and sphere test are faster than accurate testing.
 '##PD Origin Origin of the check line
 '##PD EndPos End of the check line
 '##PD CheckType You choose between 3 types of collision detection : Sphere, Box or Accurate (triangle) tests
   
 
 'We have to transform the origin and dest position with mesh matrix
   Dim TDest As D3DVECTOR, TOrigin As D3DVECTOR
 TOrigin = MITransform(Mesh(MeshIndex).Matrix, Origin)
 TDest = MITransform(Mesh(MeshIndex).Matrix, EndPos)
 
 D3DXVec3Subtract TDest, TDest, TOrigin
 D3DXVec3Normalize TDest, TDest
  
  If CheckType = TV_TESTTYPE_BOUNDINGBOX Then
    If Mesh(MeshIndex).Features.Settings.SphereComputed = False Then
     ComputeSphere
    End If
    D3DXVec3Scale TDest, TDest, 4000
    Collision = D3DX.BoxBoundProbe(Mesh(MeshIndex).Features.Settings.BoxMin, Mesh(MeshIndex).Features.Settings.BoxMax, TOrigin, TDest)
  End If
  If CheckType = TV_TESTTYPE_BOUNDINGSPHERE Then
    If Mesh(MeshIndex).Features.Settings.SphereComputed = False Then
     ComputeSphere
    End If
    Collision = D3DX.SphereBoundProbe(Mesh(MeshIndex).Features.Settings.SphereClip, Mesh(MeshIndex).Features.Settings.SphereRadius, TOrigin, TDest)
  End If
  If CheckType = TV_TESTTYPE_ACCURATETESTING Then
    Collision = DetectCollision(TOrigin, TDest, Vector(0, 0, 0))
  End If
End Function

Sub ComputeNormals()
     '##BD Computes the normal components of the vectors especially for dynamic lightning . Use this if you haven't set the nx,ny,nz arguments in the AddVertex method, or if you are using AddWall,AddFloor,.. methods, you should call this method after the creation in order to perform the dynamic lighting.
     Dim V1 As D3DVECTOR, V2 As D3DVECTOR
     Dim i As Long, E As Long
     
     If Mesh(MeshIndex).Features.Settings.VertexBufferEnabled = True Then
        D3DX.ComputeNormals Mesh(MeshIndex).D3DMesh
     Else
       With Mesh(MeshIndex).Features
          For i = 1 To .NumGroup
           If .PrimType = TV_TRIANGLESTRIP Then
             E = 0
            'For e = 0 To .Groups(i).IndicesCount - 3
             V1 = VSubtract(Vert(.Vertex(.Groups(i).VertexOffset + E + 1)), Vert(.Vertex(.Groups(i).VertexOffset + E)))
             V2 = VSubtract(Vert(.Vertex(.Groups(i).VertexOffset + E + 2)), Vert(.Vertex(.Groups(i).VertexOffset + E)))
             D3DXVec3Cross V1, V1, V2
             D3DXVec3Normalize V1, V1
           ' End If
             Debug.Print V1.x, V1.y, V1.z
             
            For E = 0 To .Groups(i).IndicesCount - 1
              DXCopyMemory .Vertex(.Groups(i).VertexOffset + E).nx, V1, 12
              'DXCopyMemory .Vertex(.Groups(i).VertexOffset + e + 1).nx, V1, 12
              'DXCopyMemory .Vertex(.Groups(i).VertexOffset + e + 2).nx, V1, 12
            Next E
           Else
            For E = 0 To .Groups(i).IndicesCount - 1 Step 3
               V1 = VSubtract(Vert(.Vertex(.Groups(i).VertexOffset + E + 1)), Vert(.Vertex(.Groups(i).VertexOffset + E)))
               V2 = VSubtract(Vert(.Vertex(.Groups(i).VertexOffset + E + 2)), Vert(.Vertex(.Groups(i).VertexOffset + E)))
               D3DXVec3Cross V1, V1, V2
               DXCopyMemory .Vertex(.Groups(i).VertexOffset + E).nx, V1, 12
               DXCopyMemory .Vertex(.Groups(i).VertexOffset + E + 1).nx, V1, 12
               DXCopyMemory .Vertex(.Groups(i).VertexOffset + E + 2).nx, V1, 12
            Next E
           End If
          Next i
       End With
     End If
End Sub

Sub FlipBillboard(LeftRight As Boolean, TopBottom As Boolean)
   '##BD Change the texture orientation of a billboard.
   '##PD LeftRight Mirror on the vertical axis
   '##PD TopBottom Mirror on the horizontal axis
    Dim temp As Single
    
    If LeftRight = True Then
      With Mesh(MeshIndex).Features
        temp = .LVertex(0).tu1
        temp = .LVertex(1).tu1
        .LVertex(0).tu1 = .LVertex(2).tu1
        .LVertex(1).tu1 = .LVertex(3).tu1
        .LVertex(2).tu1 = temp
        .LVertex(3).tu1 = temp
        temp = .Vertex(0).tu1
        temp = .Vertex(1).tu1
        .Vertex(0).tu1 = .Vertex(2).tu1
        .Vertex(1).tu1 = .Vertex(3).tu1
        .Vertex(2).tu1 = temp
        .Vertex(3).tu1 = temp
      
      End With
    End If
    If TopBottom = True Then
      With Mesh(MeshIndex).Features
        temp = .LVertex(0).tv1
        temp = .LVertex(2).tv1
        .LVertex(0).tv1 = .LVertex(1).tv1
        .LVertex(2).tv1 = .LVertex(3).tv1
        .LVertex(1).tv1 = temp
        .LVertex(3).tv1 = temp
        temp = .Vertex(0).tv1
        temp = .Vertex(1).tv1
        .Vertex(0).tv1 = .Vertex(1).tv1
        .Vertex(2).tv1 = .Vertex(3).tv1
        .Vertex(1).tv1 = temp
        .Vertex(3).tv1 = temp
      End With
    End If


End Sub

Public Function GetMatrix() As D3DMATRIX
    '##BD Returns the current matrix of the object. Only for advenced users.
    
    GetMatrix = Mesh(MeshIndex).Matrix
End Function

Sub InitLOD(Optional MinVertex As Long = 1)
   '##BD Transform this mesh to LOD mesh. You won't be able to change it later. Use this function only for big models.
    
   If Mesh(MeshIndex).Features.Settings.VertexBufferEnabled = False Then
      'Force Vertex buffer
      Optimize
      Mesh(MeshIndex).Features.Settings.VertexBufferEnabled = True
   End If
   
   ' Generate adjacent
   If Mesh(MeshIndex).Adj Is Nothing Then
    Set Mesh(MeshIndex).Adj = D3DX.CreateBuffer(Mesh(MeshIndex).D3DMesh.GetNumFaces * 12#)
    Mesh(MeshIndex).D3DMesh.GenerateAdjacency 0, ByVal Mesh(MeshIndex).Adj.GetBufferPointer
   End If
   ' Valid the mesh
   'On Error Resume Next
   If D3DX.ValidMesh(Mesh(MeshIndex).D3DMesh, ByVal Mesh(MeshIndex).Adj.GetBufferPointer) = False Then
     
     'Succeed to init LOD
     Set Mesh(MeshIndex).ProgMesh = D3DX.GeneratePMesh(Mesh(MeshIndex).D3DMesh, ByVal Mesh(MeshIndex).Adj.GetBufferPointer, ByVal 0, ByVal 0, MinVertex, D3DXMESHSIMP_VERTEX)
     Mesh(MeshIndex).Features.Settings.ProgressiveMesh = True
     Mesh(MeshIndex).ProgMesh.SetNumVertices &HFFFFFFFF
   End If
   
   
End Sub
Function IsVisible() As Integer
  '##BD Check the current visibility of the mesh according to the Sphere culling
   IsVisible = Mesh(MeshIndex).Features.Settings.Visible
End Function

Public Sub LoadDXE(Filename As String)
  '##BD Load a DXE file in the current mesh
  '##PD Filename DXE file to load.
   Dim nbface As Integer, nbfaces As String
   SetGroupPointer 1
   Dim temp As String
   Open Filename For Input As 1
     Line Input #1, temp
     Line Input #1, nbfaces
     nbface = val(nbfaces)
   Dim i As Long, x As Single, y As Single, z As Single, tu As Single, tv As Single
   Dim nx As Single, ny As Single, nz As Single
   For i = 1 To nbface * 3
      Line Input #1, temp
     x = CSng(temp)
      Line Input #1, temp
     y = CSng(temp)
     Line Input #1, temp
     z = CSng(temp)
     Line Input #1, temp
     nx = CSng(temp)
     Line Input #1, temp
     ny = CSng(temp)
     Line Input #1, temp
     nz = CSng(temp)
     Line Input #1, temp
     tu = CSng(temp)
     Line Input #1, temp
     tv = CSng(temp)
   AddVertex x, y, z, RGBA(1, 1, 1, 1), tu, tv, nx, ny, nz, False
   Next i
   Close #1
   
End Sub

Sub AddChild(ChildMesh As Mesh8)
  '##BD Add a child mesh to this mesh. The position, the rotation and the scale values will be relative of the child will be relative to this mesh, so this Mesh becomes like a "D3DRMFrame".
  '##PD Mesh Child Mesh to add
  With Mesh(MeshIndex)
     .NumChild = .NumChild + 1
     ReDim Preserve .Childs(.NumChild)
     .Childs(.NumChild) = ChildMesh
  End With
  Mesh(ChildMesh.GetMeshIndex).Parent = MeshIndex
End Sub
Function AddVertex(x, y, z, color, tu As Single, tv As Single, Optional nx As Single = 0, Optional ny As Single = 0, Optional nz As Single = 1, Optional UseTLVertex As Boolean) As Long
  '##BD Add a simple vertex, lit vertex or transformed-lit vertex to the mesh. You have to set the mesh, primitive, vertex type before calling this function. _
  If you use a COLOR_VERTEX, you should fill in the x,y,z,color,tu,tv arguments. If you use a NORMAL_VERTEX or a NON_LIT_VERTEX you should fill x,y,z,tu,tv,nx,ny,nz arguments.
 
 '##PD x X-coordinate in the 3D world
 '##PD y Y-coordinate in the 3D world
 '##PD z Z-coordinate in the 3D world
 '##PD Color Vertex color (get the color with the RGBA(r,g,b,a) function). <b>Only for LVERTEX and TLVERTEX</b>
 '##PD TU U texture coordinate
 '##PD TV V texture coordinate
 '##PD nx X-coordinate of the normal vector (only for simple Vertex)
 '##PD ny Y-coordinate of the normal vector (only for simple Vertex)
 '##PD nz Z-coordinate of the normal vector (only for simple Vertex)
 '##PD UseTLVertex Decides if the vertex will be a transformed lit vertex.
  
  Dim f As Long
  On Error GoTo Change
  'If UseTLVertex = True Then
  '  With Mesh(MeshIndex)
  '   ReDim Preserve .TLVertex(UBound(.TLVertex) + 1)
  '   .TLVertex(UBound(.TLVertex)) = TLVertex(x, y, z, color, TU, TV)
  '   AddVertex = UBound(.TLVertex)
  ' End With
  'Else
   With Mesh(MeshIndex).Features
     .NbVertex = .NbVertex + 1
    ReDim Preserve .Vertex(.NbVertex)
    .Vertex(.NbVertex - 1) = Vertex2(x, y, z, nx, ny, nz, tu, tv, tu, tv)
    ReDim Preserve .LVertex(.NbVertex)
    .LVertex(.NbVertex - 1) = LVertex2(x, y, z, color, tu, tv, tu, tv)
   ' Debug.Print x, y, z
    If GroupPointer > -1 Then
      f = .Groups(GroupPointer).IndicesCount + 1
      ReDim Preserve .Groups(GroupPointer).Indices(f)
      .Groups(GroupPointer).Indices(f) = .NbVertex - 1
      .Groups(GroupPointer).IndicesCount = f
    
    End If
    AddVertex = .NbVertex
  End With
' End If
 Exit Function
Change:
 Debug.Print "err :"; Err.Description
End Function
Public Sub AddWall2(Texture As Texture8, x1 As Single, z1 As Single, X2 As Single, z2 As Single, height1 As Single, height2 As Single, Optional Altitude1 As Single = 0, Optional Altitude2 As Single = 0, Optional TileW As Single = 1, Optional TileH As Single = 1)
    'Tile created by GabitaSoft Belguim (www.Gabita.com or www.VisualBasicTec.com)
    '##BD This second Wall method allows you to create easily textured walls. Rooms that are created with this method can be lit by the GenerateLightMap. Coordinates are the same that for a line.
    '##PD x1 X1 Wall coordinate
    '##PD z1 Y1 Wall coordinate
    '##PD X2 X2 Wall coordinate
    '##PD z2 Y2 Wall coordinate
    '##PD height1 Wall height 1
    '##PD height2 Wall height 2
    '##PD Altitude1 Wall Altitude 1.
    '##PD Altitude2 Wall Altitude 2
    AddFace Texture, True
    AddVertex x1, Altitude1, z1, D3DColorMake(1, 1, 1, 1), 0.02 * TileH, 0.02 * TileW, 0, 0, 0
    AddVertex x1, height1 + Altitude1, z1, D3DColorMake(1, 1, 1, 1), 0.02 * TileH, 0.98 * TileW, 0, 0, 0
    AddVertex X2, Altitude2, z2, D3DColorMake(1, 1, 1, 1), 0.98 * TileH, 0.02 * TileW, 0, 0, 0
    AddVertex X2, height2 + Altitude2, z2, D3DColorMake(1, 1, 1, 1), 0.98 * TileH, 0.98 * TileW, 0, 0, 0
 End Sub
Sub AddFace(Texture As Texture8, lightmap As Boolean)
  '##BD Add a face (and a group) to the current mesh.
  '##PD Texture Texture class that will be applied on the face.
  '##PD LightMap Decides if this face will use a lightmap.
  Dim Grp As Integer
  SetGroupPointer Mesh(MeshIndex).Features.NumGroup + 1
  Grp = Mesh(MeshIndex).Features.NumGroup
  If Not Texture Is Nothing Then
    Mesh(MeshIndex).Features.Groups(Grp).TextureI = Texture.GetTexIndex
  Else
    Mesh(MeshIndex).Features.Groups(Grp).TextureI = 0
  End If
  
  Mesh(MeshIndex).Features.Groups(Grp).LightmapI = IIf(lightmap = True, 0, -1)
End Sub

Public Sub AddFaceFromPoint(ByVal Texture As Texture8, x1 As Single, y1 As Single, z1 As Single, X2 As Single, y2 As Single, z2 As Single, X3 As Single, Y3 As Single, Z3 As Single, X4 As Single, Y4 As Single, Z4 As Single, Optional TileW As Single = 1, Optional TileH As Single = 1)
    '##BD Add a face as a new mesh group from 4 given points.
    '##PD TileW  Texture Tiling on the horizontal axis
    '##PD TileH  Texture Tiling on the vertical axis
    'Tile created by GabitaSoft Belguim (www.Gabita.com or www.VisualBasicTec.com)
    'Created by De Keyser Sigurd (a programmer from GabitaSoft Belgium)
    ' www.Gabita.com
     SetPrimitiveType TV_TRIANGLESTRIP
    AddFace Texture, True
    AddVertex x1, y1, z1, D3DColorMake(1, 1, 1, 1), 0.02 * TileH, 0.02 * TileW, 1, 0, 0
    AddVertex X2, y2, z2, D3DColorMake(1, 1, 1, 1), 0.02 * TileH, 0.98 * TileW, 1, 0, 0
    AddVertex X3, Y3, Z3, D3DColorMake(1, 1, 1, 1), 0.98 * TileH, 0.02 * TileW, 1, 0, 0
    AddVertex X4, Y4, Z4, D3DColorMake(1, 1, 1, 1), 0.98 * TileH, 0.98 * TileW, 1, 0, 0

End Sub

Public Sub AddTriangle(ByVal Texture As Texture8, x1 As Single, y1 As Single, z1 As Single, X2 As Single, y2 As Single, z2 As Single, X3 As Single, Y3 As Single, Z3 As Single, Optional TileW As Single = 1, Optional TileH As Single = 1)
    '##BD Add a triangle as a new mesh group from 3 given points.
    '##PD TileW  Texture Tiling on the horizontal axis
    '##PD TileH  Texture Tiling on the vertical axis
   
   'Tile created by GabitaSoft Belguim (www.Gabita.com or www.VisualBasicTec.com)
    'Created by De Keyser Sigurd (a programmer from GabitaSoft Belgium)
    ' www.Gabita.com
    AddFace Texture, True
    AddVertex x1, y1, z1, D3DColorMake(1, 1, 1, 1), 0.02 * TileH, 0.02 * TileW, 1, 0, 0
    AddVertex X2, y2, z2, D3DColorMake(1, 1, 1, 1), 0.02 * TileH, 0.98 * TileW, 1, 0, 0
    AddVertex X3, Y3, Z3, D3DColorMake(1, 1, 1, 1), 0.98 * TileH, 0.02 * TileW, 1, 0, 0
End Sub




Public Sub AddWall(Texture As Texture8, x1 As Single, z1 As Single, X2 As Single, z2 As Single, height As Single, Optional Altitude As Single = 0, Optional TileW As Single = 1, Optional TileH As Single = 1)
    'Tile created by GabitaSoft Belguim (www.Gabita.com or www.VisualBasicTec.com)
    '##BD This simple method allows you to create easily textured walls. Rooms that are created with this method can be lit by the GenerateLightMap. Coordinates are the same that for a line.
    '##PD x1 X1 Wall coordinate
    '##PD z1 Y1 Wall coordinate
    '##PD X2 X2 Wall coordinate
    '##PD z2 Y2 Wall coordinate
    '##PD height Wall height
    '##PD Altitude Wall Altitude.
    AddFace Texture, True
    AddVertex x1, Altitude, z1, RGBA(1, 1, 1, 1), 0.02 * TileH, 0.02 * TileW, 1, 0, 0
    AddVertex x1, height + Altitude, z1, RGBA(1, 1, 1, 1), 0.02 * TileH, 0.98 * TileW, 1, 0, 0
    AddVertex X2, Altitude, z2, RGBA(1, 1, 1, 1), 0.98 * TileH, 0.02 * TileW, 0, 0, 0
    AddVertex X2, height + Altitude, z2, RGBA(1, 1, 1, 1), 0.98 * TileH, 0.98 * TileW, 1, 0, 0
    SetPrimitiveType TV_TRIANGLESTRIP
End Sub
Public Sub AddFloor(Texture As Texture8, x1 As Single, z1 As Single, X2 As Single, z2 As Single, Optional Altitude As Single = 0, Optional TileW As Single = 1, Optional TileH As Single = 1)
    'Tile created by GabitaSoft Belguim (www.Gabita.com or www.VisualBasicTec.com)
    '##BD A simple method that allows you to create easily a floor face. The floor can also be used for a ceiling.
    '##PD Texture Texture class that will be applied on the floor.
    '##PD x1 Left coordinate of the face
    '##PD z1 Top coordinate of the face
    '##PD X2 Right coordinate of the face
    '##PD z2 Bottom coordinate of the face
    '##PD Altitude Floor (or ceiling) altitude.
    AddFace Texture, True
    SetPrimitiveType CONST_TV_PRIMITIVETYPE.TV_TRIANGLESTRIP
    AddVertex x1, Altitude, z1, D3DColorMake(1, 1, 1, 1), 0.02 * TileH, 0.02 * TileW, 0, -1, 0
    AddVertex x1, Altitude, z2, D3DColorMake(1, 1, 1, 1), 0.02 * TileH, 0.98 * TileW, 0, -1, 0
    AddVertex X2, Altitude, z1, D3DColorMake(1, 1, 1, 1), 0.98 * TileH, 0.02 * TileW, 0, -1, 0
    AddVertex X2, Altitude, z2, D3DColorMake(1, 1, 1, 1), 0.98 * TileH, 0.98 * TileW, 0, -1, 0
End Sub

Public Sub ComputeSphere()
   '##BD Computes the Bounding Sphere. You need to call this function if you want the object be culled to improve the speed of the application. Note that you should use this function after the entire creation of the mesh.
   
   With Mesh(MeshIndex).Features
    If .NbVertex <> 0 Then
      If .Settings.VertexBufferEnabled = True Then
        D3DX.ComputeBoundingSphereFromMesh Mesh(MeshIndex).D3DMesh, .Settings.SphereClip, .Settings.SphereRadius
        D3DX.ComputeBoundingBoxFromMesh Mesh(MeshIndex).D3DMesh, .Settings.BoxMin, .Settings.BoxMax
      Else
        D3DX.ComputeBoundingSphere .Vertex(0), .NbVertex + 1, TV_VERTEX2, .Settings.SphereClip, .Settings.SphereRadius
        D3DX.ComputeBoundingBox .Vertex(0), .NbVertex + 1, TV_VERTEX2, .Settings.BoxMin, .Settings.BoxMax
        
      End If
       With .Settings
       .BoxMin = VMult(.BoxMin, .Scaling)
       .BoxMax = VMult(.BoxMax, .Scaling)
       .SphereClip = VMult(.SphereClip, .Scaling)
       .SphereRadius = .SphereRadius * VMaX(.Scaling)
       
       .SphereComputed = True
       End With
    End If
   End With

End Sub

Public Sub Create3DText(Text As String, Optional UserFont As String = "")
   '##BD Doesn't work at the moment
 
   If UserFont = "" Then
     D3DX.CreateText Device, LM.hdc, Text, 1, 40, Mesh(MeshIndex).D3DMesh, ByVal 0
     Mesh(MeshIndex).Created = True
     Mesh(MeshIndex).Features.Settings.VertexBufferEnabled = True
     ScaleMesh 4, 4, 4
   End If
End Sub
Function CreateChild(Optional Name As String = "") As Mesh8
   '##BD Creates and returns a new mesh as a child of the current mesh.
   '##PD Name Name of the new mesh.
   Set CreateChild = LocalScene.CreateMeshBuilder(Name)
    With Mesh(MeshIndex)
     .NumChild = .NumChild + 1
     ReDim Preserve .Childs(.NumChild)
     .Childs(.NumChild) = CreateChild.GetMeshIndex
  End With
  Mesh(CreateChild.GetMeshIndex).Parent = MeshIndex
End Function

Public Sub CreateSphere(Texture As Texture8, Radius As Single, Slice As Long, Stack As Long, Optional color As Long = &HFFFFFFFF)
   '##BD Create a sphere from three parameters.
   '##PD Texture texture that will be applied on the sphere.
   '##PD Radius Radius of the sphere
   '##PD Slice Number of slices of the sphere. More this number is big, more the sphere is accurate.
   '##PD Stack Number of "meridians" of the sphere. More this number is big, more the sphere is accurate.
   SetTexture Texture
   Set Mesh(MeshIndex).D3DMesh = D3DX.CreateSphere(Device, Radius, Slice, Stack, Nothing)
   Mesh(MeshIndex).Features.Settings.VertexBufferEnabled = True
   Mesh(MeshIndex).Features.NbVertex = Mesh(MeshIndex).D3DMesh.GetNumVertices
   SetGroupPointer 1
   Mesh(MeshIndex).Features.Groups(1).IndicesCount = 30
End Sub


Sub CreateTeapot()
   '##BD Create a teapot from nothing. This method can only be used to test the engine.
   SetGroupPointer 1
   Set Mesh(MeshIndex).D3DMesh = D3DX.CreateTeapot(Device, Nothing)
   Mesh(MeshIndex).Created = True
   Mesh(MeshIndex).Features.Settings.VertexBufferEnabled = True
   ScaleMesh 4, 4, 4
   Mesh(MeshIndex).Changed = True
   Mesh(MeshIndex).Features.NbVertex = Mesh(MeshIndex).D3DMesh.GetNumVertices
   Mesh(MeshIndex).Features.Groups(1).IndicesCount = 30

End Sub

Sub DeleteChilds()
      '##BD Delete the childs of the current mesh.
      Dim i As Long
      For i = 1 To Mesh(MeshIndex).NumChild
        Mesh(Mesh(MeshIndex).Childs(i)).Parent = 0
      Next i
      Mesh(MeshIndex).NumChild = 0
      ReDim Mesh(MeshIndex).Childs(0)
End Sub

Private Function DetectCollision(Origin As D3DVECTOR, dest As D3DVECTOR, normal As D3DVECTOR, Optional ByRef U As Single, Optional ByRef V As Single, Optional ByRef t As Single)
 '##BD Check if the line between the origin and the dest point intersects a triangle of this mesh.
 '##PD Origin Start point of the line (usually the player position)
 '##PD dest End point of the line (usually the player position + player direction)
 '##PD Normal Returns the collide face normal.
 '##PD U Returns the U-coordinate on the collide face
 '##PD V Returns the V-coordinate on the collide face
 '##PD t Returns the distance between the origin and the intersection point.
 
 On Error Resume Next
   With Mesh(MeshIndex)
    'get the ray direction
  Dim x As Long, g As Long, i As Long
       
 
 
'    Dim T As Single
   'multi group mesh
    
    For i = 1 To .Features.NumGroup
     For x = 0 To .Features.Groups(i).IndicesCount - 3
      'g = RayTri(Vert(.Features.Vertex(.Features.Groups(i).VertexOffset + x)), Vert(.Features.Vertex(.Features.Groups(i).VertexOffset + x + 1)), Vert(.Features.Vertex(.Features.Groups(i).VertexOffset + x + 2)), Dest, Origin, t, 0, 0, True)
      g = Ray_Tri(Origin, dest, Vert(.Features.Vertex(.Features.Groups(i).VertexOffset + x)), Vert(.Features.Vertex(.Features.Groups(i).VertexOffset + x + 1)), Vert(.Features.Vertex(.Features.Groups(i).VertexOffset + x + 2)), t, 0, 0, True)
      
      If (g = 1 And t < PrecisionCollide And t >= 0) Then
          DetectCollision = True
          normal = CollisionNormal
          Exit Function
      End If
    Next x
  Next i
  End With
  Exit Function
GestE:
  'AddLog "DetectCollision " + CStr(MeshIndex) + " " + Err.Description + " " + CStr(Err.Number)
  Resume Next
  
End Function


Sub Enable(OnOff As Boolean)
     '##BD Enable/Disable the mesh. If a parent mesh is disable, then its  child meshes are disable too.
     '##PD OnOff Enable/Disable
     Mesh(MeshIndex).Enabled = OnOff
End Sub

Sub EnableSphereMapping(OnOff As Boolean, Optional Mode As Long = 0)
   Mesh(MeshIndex).Features.Settings.SphereMapping = OnOff
   If Mode = 0 Then
   Mesh(MeshIndex).Features.Settings.SphereMappingMode = D3DTSS_TCI_CAMERASPACEREFLECTIONVECTOR
   Else
   Mesh(MeshIndex).Features.Settings.SphereMappingMode = D3DTSS_TCI_CAMERASPACENORMAL
   End If
End Sub

Private Sub GenerateVertexBuffer()
    With Mesh(MeshIndex)
      
    End With
    
End Sub

Sub Load3DsMesh(Filename As String)
  '##BD Load a 3DS file in the mesh. The materials and the texture will be loaded too. The textures MUST be BMP,GIF,JPG or DDS and have to be in the same directory of the model.
  '##PD filename 3Ds filename.
    Dim LE As MaterialFactory8
    Set LE = New MaterialFactory8

    'SetPrimitiveType TV_TRIANGLELIST
    'SetVertexType TV_VERTEX
    On Error GoTo errs
    Dim ErrString As String
    ErrString = "3DS: Load3ds"
   ' AddLog "Load 3ds :" + filename
    Dim VbCounter As Long
    Dim ColorByte As udColorByte
    Dim PercentInt As Integer
    Dim cBlock As Byte              ' Type of color currently being imported. See colorblock Enum
    Dim Materials2() As aMaterial
    Dim mCount As Integer           ' Number of loaded material
    'im MatTexture() As Texture8
    
    Dim Info As tdsHeader           ' Header of Chunk
    Dim Index As Long               ' Index in File -- filepos
        
    Dim MeshName As String          ' Current Mesh Name
    Dim Done As Boolean
    Dim CurrentMat As Integer
    Dim NbGroupe As Integer
    Dim vcount As Long              ' redundant vertex counter
    Dim vertices As Integer         ' ""
    Dim faces As Integer            ' Numbder of faces
    Dim VertexList() As D3DVECTOR   ' #DS file Vertexlist
    Dim Facelist() As aFace         ' Facelist
    Dim vbuf() As Long              ' redunant vertexlist for vertex buffer
    'Dim MeshVertices() As D3DVERTEX ' Actual constructor for vertex buffer
    
    Dim UVCoords() As UVmap         ' #DS file UV coordinates
    Dim UVbuf() As UVmap            ' Buffer for MeshVertices construction
    Dim nBuf() As D3DVECTOR         ' Buffer for the temp face Normals
    Dim Normals() As D3DVECTOR      ' Array to hold temporary face normals
    
    Dim tVector As D3DVECTOR
    Dim Char As Byte
    Dim Strz As String              ' All Generic definitions for reading
    Dim Dword As Long               ' types stored in #DS file via C typedefs
    Dim word As Integer
    Dim Real As Single
    Dim i As Long
    Dim TheScale As Long
     Mesh(MeshIndex).Features.PrimType = D3DPT_TRIANGLELIST
     TheScale = 1
    Dim UVCoordsTrue As Boolean
    UVCoordsTrue = False
        
    Dim MinX, MinY, MinZ As Single
    Dim MaxX, MaxY, maxZ As Single
    
    Index = 1

    TestPAKFile Filename
    AddLog "Loading 3DS mesh : " + CStr(Filename) + "..."
    Open Filename For Binary As #1
    
 
    While Not EOF(1)
        Get #1, Index, Info
        Index = Index + 6
'        Debug.Print Hex(Info.Header), Info.Length
        Select Case Info.Header
                
            Case &H2:       ' 3DS Version Number
                ErrString = "3DS Version"
                Get #1, Index, Dword
                Index = Index + Len(Dword)
                
            Case &H3D3D:    ' 3D Editor Header
                ErrString = "3D Editor Chunk"
                
            Case &H3D3E:    ' Mesh Version
                ErrString = "Mesh Version"
                Get #1, Index, Dword
                Index = Index + Len(Dword)
                
            Case &H4000:    ' Object Block -- Includes Name
                ErrString = "Object Block"
                Done = False
                MeshName = ""
                While Not Done
                    Get #1, Index, Char
                    If Char <> 0 Then
                        MeshName = MeshName + Chr(Char)
                    Else
                        Done = True
                    End If
                    Index = Index + Len(Char)
                Wend
                Debug.Print "Name :"; MeshName; ""
                
            Case &H4100:    ' Triangular Mesh
                ErrString = "Triangular Mesh"
                NbGroupe = NbGroupe + 1
                SetGroupPointer NbGroupe
                Debug.Print NbGroupe
            Case &H4110:    ' Verticies List
                ErrString = "Vertices"
                Get #1, Index, vertices
                Index = Index + Len(vertices)
                Debug.Print "Vertex list "; vertices
                ReDim VertexList(vertices)
                Get #1, Index, VertexList
                Index = Index + Len(VertexList(0)) * CLng(vertices)
                
                MinX = 500000:  MinY = MinX: MinZ = MinY
                MaxX = -500000: MaxY = MaxX: maxZ = MaxY
                  
                For i = 0 To vertices - 1
                    With VertexList(i)
                        VertexList(i).x = VertexList(i).x * TheScale
                        VertexList(i).y = VertexList(i).y * TheScale
                        VertexList(i).z = VertexList(i).z * TheScale
                        If .x < MinX Then MinX = .x
                        If .y < MinY Then MinY = .y
                        If .z < MinZ Then MinZ = .z
                    
                        If .x > MaxX Then MaxX = .x
                        If .y > MaxY Then MaxY = .y
                        If .z > maxZ Then maxZ = .z
                    End With
                Next i
              
            Case &H4120:    ' Face List
                ErrString = "Face List"
                Get #1, Index, faces
                Index = Index + Len(faces)
                Debug.Print "Face list "; faces
                
                ReDim Facelist(faces)
                Get #1, Index, Facelist
                Index = Index + Len(Facelist(0)) * CLng(faces)
            
                vcount = -1
                ReDim vbuf(faces * 3)
                ReDim UVbuf(faces * 3)
                For i = 0 To faces - 1
                    vcount = vcount + 3
                    'debug.print vcount
                    If UVCoordsTrue Then
                        UVbuf(vcount - 2) = UVCoords(Facelist(i).a)
                        UVbuf(vcount - 1) = UVCoords(Facelist(i).b)
                        UVbuf(vcount) = UVCoords(Facelist(i).C)
                    End If
                    vbuf(vcount - 2) = Facelist(i).a
                    vbuf(vcount - 1) = Facelist(i).b
                    vbuf(vcount) = Facelist(i).C
                    
                    ' Make Face Normal
                    Dim a As D3DVECTOR, b As D3DVECTOR, C As D3DVECTOR
                    ReDim Preserve nBuf(vcount)
                    With DirectX
                        ' Face Normal
                    
                        D3DXVec3Subtract a, VertexList(vbuf(vcount - 2)), VertexList(vbuf(vcount - 1))
                        D3DXVec3Subtract b, VertexList(vbuf(vcount - 2)), VertexList(vbuf(vcount))
                        D3DXVec3Cross C, a, b
                        D3DXVec3Normalize C, C
                        
                        nBuf(vcount) = C
                        nBuf(vcount - 1) = C
                        nBuf(vcount - 2) = C
                        
                    End With
                Next i
                UVCoordsTrue = False
                
                       
                ErrString = "Load3ds: Mesh Add"
               Debug.Print "Mesh add " + MeshName
                With Mesh(MeshIndex)
                     
                    For i = 0 To vcount
                      'AddLog "Addvertex :" + CStr(vbuf(I))
                        AddVertex VertexList(vbuf(i)).x, VertexList(vbuf(i)).y, VertexList(vbuf(i)).z, D3DColorMake(1, 1, 1, 1), UVbuf(i).U, 1 - UVbuf(i).V, nBuf(i).x, nBuf(i).y, nBuf(i).z
                    Next i
                End With
                
            Case &H4130:    ' Face Material List
                ErrString = "Face Material"
                Done = False
                Strz = ""
                While Not Done
                    Get #1, Index, Char
                    If Char <> 0 Then
                        Strz = Strz + Chr(Char)
                    Else
                        Done = True
                    End If
                    Index = Index + Len(Char)
                Wend
                
                Dim Q As Integer
                For Q = 1 To mCount
                    If Strz = Materials2(Q).MatName Then
                        ' Store the Material Here
                         Debug.Print "Selected :"; Strz
                         Mesh(MeshIndex).Features.Groups(NbGroupe).TextureI = Materials2(Q).MapPercent
                         Materials(Materials2(Q).MaterialIndex).Mat.Ambient = Materials2(Q).Ambient
                         Materials(Materials2(Q).MaterialIndex).Mat.Diffuse = DXColor(1, 1, 1, 1)
                         Materials(Materials2(Q).MaterialIndex).Mat.Specular = Materials2(Q).Specular
                         Materials(Materials2(Q).MaterialIndex).Mat.emissive = DXColor(0, 0, 0, 0)
                         'Debug.Print Q, Materials2(Q).Ambient.r, , Materials2(Q).Ambient.g, Materials2(Q).Ambient.B
                         Mesh(MeshIndex).Features.Groups(NbGroupe).MaterialI = Materials2(Q).MaterialIndex
                         'SetTexture MatTexture(Q), NbGroupe
                         Exit For
                    End If
                Next Q
                
                Get #1, Index, word
                Index = Index + Len(word)
                For i = 1 To word
                    Get #1, Index, word
                    Index = Index + Len(word)
                Next i
                
            Case &H4140:    ' Face Texture Coordinate
                Debug.Print "Add texture coordinates"
                ErrString = "Face Texture Coordinates"
                Get #1, Index, word
                Index = Index + Len(word)
                ReDim UVCoords(word)
                Get #1, Index, UVCoords
                Index = Index + Len(UVCoords(0)) * CLng(word)
                UVCoordsTrue = True
            
            Case &H4160:
                ErrString = "Local Coordinate System"
                Get #1, Index, tVector
                Index = Index + Len(tVector)
                Get #1, Index, tVector
                Index = Index + Len(tVector)
                Get #1, Index, tVector
                Index = Index + Len(tVector)
                Get #1, Index, tVector
                'DD.MeshParent(ObjName2).Mesh(MeshName).position
                Index = Index + Len(tVector)
            
            Case &HA000:    ' Material Name
                ErrString = "Material Name"
                Done = False
                Strz = ""
                While Not Done
                    Get #1, Index, Char
                    If Char <> 0 Then
                        Strz = Strz + Chr(Char)
                    Else
                        Done = True
                    End If
                    Index = Index + Len(Char)
                Wend
                Materials2(mCount).MatName = Strz
                Debug.Print "Material Name :"; Strz
            Case &HAFFF:    ' Material Editor Chunk
                ErrString = "Material Editor Chunk"
                mCount = mCount + 1
                ReDim Preserve Materials2(mCount)        ' Save stored materials
                CurrentMat = LE.CreateMaterial
                Materials2(mCount).MaterialIndex = CurrentMat
            Case &H10:      ' Single Color Chunk
                ErrString = "Single Color Chunk"
                Index = Index + Info.Length - 6
                
            Case &H11:      ' Byte Color Chunk
                ErrString = "Byte Color Chunk"
                Get #1, Index, ColorByte
                Index = Index + Len(ColorByte)
                
                Select Case cBlock
                    Case Ambient:
                        With Materials2(mCount).Ambient
                            .a = 1
                            .r = ColorByte.Red / 255
                            .g = ColorByte.Green / 255
                            .b = ColorByte.Blue / 255
                        Debug.Print " Material Ambient :"; .r; " "; .g; " "; .b
                        End With
                    Case Diffuse:
                        With Materials2(mCount).Diffuse
                            .a = 1
                            .r = ColorByte.Red / 255
                            .g = ColorByte.Green / 255
                            .b = ColorByte.Blue / 255
                        Debug.Print " Material Diffuse :"; .r; " "; .g; " "; .b
                        
                        End With
                        
                    Case Specular:
                        With Materials2(mCount).Specular
                            .a = 1
                            .r = ColorByte.Red / 255
                            .g = ColorByte.Green / 255
                            .b = ColorByte.Blue / 255
                        Debug.Print " Material Specular :"; .r; " "; .g; " "; .b
                        
                        End With
                End Select
                
            Case &H30:      ' Percent Chunk Int format
                ErrString = "Percent Chunk INT"
                Get #1, Index, PercentInt
                Index = Index + Len(PercentInt)
                
            Case &H31:      ' Percent Chunk Float format
                ErrString = "Percent Chunk Float"
                Index = Index + Info.Length - 6
            
            Case &HA010:    ' Ambient color
                ErrString = "Ambient Color"
                cBlock = Ambient
                
            Case &HA020:    ' Diffuse Color
                ErrString = "Diffuse Color"
                cBlock = Diffuse
                
            Case &HA030:    ' Specular Color
                ErrString = "Specular Color"
                cBlock = Specular
                
            Case &HA200:    ' Map #1 -- diffuse?
                cBlock = Diffuse
                
            Case &HA300:    ' Map FileName
                ErrString = "Map File Name"
                Done = False
                Strz = ""
                While Not Done
                    Get #1, Index, Char
                    If Char <> 0 Then
                        Strz = Strz + Chr(Char)
                    Else
                        Done = True
                    End If
                    Index = Index + Len(Char)
                Wend
                Select Case cBlock
                    Case Diffuse:
                        Materials2(mCount).MapName = Strz
                        
                End Select
                
                Debug.Print "Material Map Name: " & Strz
                'ReDim Preserve MatTexture(mCount)
                'For i = 1 To 256
                '  If TextureName(i) = "" Then Exit For
                'Next i
                If Exist(Strz) = False Then
                  Strz = Left$(Filename, InStrRev(Filename, "\")) + Strz
                End If
                i = LocTexFactory.LoadTexture(Strz, Strz, , , TV_COLORKEY_BLACK).GetTexIndex
                'Set Texture(i) = LoadSimpleTexture(SearchExistPath(Strz, Left$(Filename, InStrRev(Filename, "\"))), 256, 256)
                'TextureName(i) = "3DS Texture " + CStr(i)
                
                Materials2(mCount).MapPercent = i
                'Set MatTexture(mCount) = LocalScene.LoadTexture("d:\" + Strz, 256, 256)
            Case &HA33A:    ' Map #2 -- Ambient?
                cBlock = Ambient
            
            Case &HA354
                ErrString = "V scaling for Map"
                Index = Index + Info.Length - 6
                
            Case &HA356
                ErrString = "U scaling for Map"
                Index = Index + Info.Length - 6
                
            Case &HA358
                ErrString = "U offset for Map"
                Index = Index + Info.Length - 6
                
            Case &HA35A
                ErrString = "V offset for Map"
                Index = Index + Info.Length - 6
                
            Case &HA35C
                ErrString = "Rotation Angle for Map"
                Index = Index + Info.Length - 6
                
            Case &H100:     ' One Unit
                ErrString = "One Unit"
                Get #1, Index, Real
                Index = Index + Len(Real)
            Case &H4D4D ' OK the object begins
            
            Case Else       ' Unknown Chunk
                ErrString = "Unknown Chunk"
                Index = Index + Info.Length - 6
        End Select
        Debug.Print "  ##  "; Hex$(Info.Header); "("; ErrString; ")"
        AddLog "  ##  " + Hex$(Info.Header) + " (" + CStr(ErrString) + ")"
    Wend
    Close #1
    DeletePAKTemp Filename
    AddLog "End of 3DS Loading."
    Exit Sub
errs:
Debug.Print "Error :"; ErrString, Err.Number, Err.Description
AddLog "Error in 3Ds loader :" + CStr(Err.Number) + " " + Err.Description + " " + ErrString
Resume Next
End Sub

Sub GenerateLightMap(LightMapPrecision As Single, Shadow As Boolean, Optional AmbientColor As Long)
   '##BD Generate the lightmap for this mesh. This method can take several seconds.
   '##PD LightMapPrecision Precision of the lightmaps that will be created (usually 5)
   '##PD Shadow Decides if TrueVision must create shadows too (slower)
   '##PD AmbientColor AmbientColor you can create with RGBA() function
   Dim i As Integer
    Dim C(2) As Integer, V1 As D3DVECTOR, V2 As D3DVECTOR, v3 As D3DVECTOR, v4 As D3DVECTOR, V(4) As D3DVECTOR
    Dim DistV As Single, TestV As D3DVECTOR, ch As Long, cw As Long, Chh As Long, cww As Long
    Dim AmbientR As Single, AmbientB As Single, AmbientG As Single, AmbientA As Single, scalex As Single, scaley As Single
    Dim j As Long, h As Long, hdc As Long, x As Long, y As Long, a As Single, test As Single, intensity As Single
    Dim offset1 As Long, by() As Byte
   LightMapPrecision = LightMapPrecision / 16
  
If DisplayConsole = True Then
      ClearScreen
      LocalScreen.CreateUserFont "LightMap", "Tahoma", 8, True, False, False
      LocalScreen.DrawFilledColorBox 10, 5, 200, 130, &HFF777700, &HFF007700, &HFF007777, &HFF770000, False
      LocalScreen.DrawBox 10, 5, 200, 130, &HFF00FF00
      LocalScreen.DrawText "Lightmaps Generator", 20, 10, &HFFFF0000, "LightMap"
      LocalScreen.DrawText "--------------------------------", 20, 20, &HFFFF0000, "LightMap"
      LocalScreen.DrawText "Create Lightmaps...", 20, 30, &HFF00FFFF, "LightMap"
      LocalScreen.DrawText "with Shadows...", 20, 40, &HFF00FFFF, "LightMap"
      LocalScreen.DrawFilledBox 20, 100, 180, 110, &H77000000, True
      LocalScreen.DrawBox 20, 100, 180, 110, &HFF00FF00
      FlipToScreen
End If

  If AmbientColor <> 0 Then
    'AmbientR = DirectX.ColorGetRed(AmbientColor)
    'AmbientB = DirectX.ColorGetBlue(AmbientColor)
    'AmbientG = DirectX.ColorGetGreen(AmbientColor)
    'AmbientA = DirectX.ColorGetAlpha(AmbientColor)
    Dim CS As D3DCOLORVALUE
    CS = LCToDXC(AmbientColor)
    AmbientR = CS.r
    AmbientG = CS.g
    AmbientB = CS.b
    AmbientA = 1
  End If
  Dim GetNum As Single
  On Error GoTo ErrorG
  With Mesh(MeshIndex)
   If .Features.NumGroup = 0 Then AddLog "Can't do lightmapping on more than 4 vertexes": Exit Sub
   For i = 1 To .Features.NumGroup
       If DisplayConsole = True Then
          GetNum = i / .Features.NumGroup
           LocalScreen.DrawFilledBox 21, 101, 21 + 158 * (i / .Features.NumGroup), 109, RGBA(1, GetNum, 0, 1), False
          FlipToScreen
       End If

       If .Features.Groups(i).LightmapI = 0 Then
         For j = 1 To NumMeshLM
          If MeshLM(j) Is Nothing Then .Features.Groups(i).LightmapI = j: Exit Sub
         Next j
         If .Features.Groups(i).LightmapI = 0 Then
           NumMeshLM = NumMeshLM + 1
           ReDim Preserve MeshLM(NumMeshLM)
           .Features.Groups(i).LightmapI = NumMeshLM
         End If
       Else
         Set MeshLM(.Features.Groups(i).LightmapI) = Nothing
       End If
       
       'Simple lightmap generation
      For h = 0 To .Features.Groups(i).IndicesCount - 1
       V(h) = Vert(.Features.Vertex(.Features.Groups(i).VertexOffset + h))
      Next h
      V1 = VSubtract(V(2), V(0))
      V2 = VSubtract(V(1), V(0))
      cw = (VModulus(V2) * LightMapPrecision)
      ch = (VModulus(V1) * LightMapPrecision)
      cww = cw: Chh = ch
     
      'If cw < 8 Then cw = 8
      'If ch < 8 Then ch = 8
      D3DX.CheckTextureRequirements Device, cw, ch, 1, 0, D3DFMT_R5G6B5, D3DPOOL_DEFAULT
'      GetSquaredSize cw, ch
      
      Set MeshLM(.Features.Groups(i).LightmapI) = D3DX.CreateTexture(Device, cw, ch, 1, 0, D3DFMT_R5G6B5, D3DPOOL_DEFAULT)
      
      'MeshLM(.GroupIndex(i).LightMap).BltColorFill REC(0, 0, 0, 0), 0
      'hdc = MeshLM(.GroupIndex(i).LightMap).GetDC
      offset1 = 0
      ReDim Preserve by(cw * ch * 2#) As Byte
      For y = 0 To ch - 1
        For x = 0 To cw - 1
          scalex = x / cw
          scaley = y / ch
          v3 = VScale(V1, scalex)
          v4 = VScale(V2, scaley)
          TestV = VAdd(V(0), VAdd(v3, v4))
          C(0) = AmbientR * 255: C(1) = AmbientG * 255: C(2) = AmbientB * 255
          For a = 0 To 16
            If Light(a).Active = True And Light(a).Light.Type = D3DLIGHT_POINT Then
              DistV = VModulus(VSubtract(TestV, Light(a).Light.Position))
              test = 0
              If Shadow = True Then
                test = CheckFaces(Light(a).Light.Position, TestV, DistV, i)
              End If
              intensity = 1 - (DistV / Light(a).Light.range)
              If intensity > 0 And test = 0 Then
                C(0) = C(0) + Int(intensity * 255# * Light(a).Light.Diffuse.r)
                C(1) = C(1) + Int(intensity * 255# * Light(a).Light.Diffuse.g)
                C(2) = C(2) + Int(intensity * 255# * Light(a).Light.Diffuse.b)
              End If
            End If
          Next a
          If C(0) > 255 Then C(0) = 255
          If C(1) > 255 Then C(1) = 255
          If C(2) > 255 Then C(2) = 255
          by(offset1 + 1) = (C(0) And &HF8) Or (C(1) And &HE0) \ 32
          by(offset1) = (C(1) And &H1C) * 8 Or (C(2) And &HF8) \ 8
          offset1 = offset1 + 2
        Next x
      Next y
      'MeshLM(.GroupIndex(i).LightMap).ReleaseDC hdc
      Dim CSA As Direct3DSurface8
      Set CSA = Device.CreateImageSurface(cw, ch, D3DFMT_R5G6B5)
      D3DX.LoadSurfaceFromMemory CSA, ByVal 0, ByVal 0, by(0), D3DFMT_R5G6B5, cw * 2, ByVal 0, REC(0, 0, cw, ch), D3DX_DEFAULT, 0
      Device.CopyRects CSA, ByVal 0, 0, MeshLM(.Features.Groups(i).LightmapI).GetSurfaceLevel(0), ByVal 0
   
   Next i
  
  End With
 Exit Sub
ErrorG:
AddLog "Generate LightMap " + CStr(Err.Number)
   If DisplayConsole = True Then
     '    LocalConsole.WriteConsole "Lightmaps generated."
    End If

 Resume Next
End Sub
Private Function CheckFaces(StartPos As D3DVECTOR, EndPos As D3DVECTOR, DistMax As Single, CurrentFace As Integer)
   Dim TOrigin As D3DVECTOR, t As Single
   Dim TDest As D3DVECTOR, TDest2 As D3DVECTOR
   TOrigin = StartPos
   TDest = VSubtract(EndPos, TOrigin)
   D3DXVec3Normalize TDest2, TDest
 Dim i As Long, x As Long, g As Long
 With Mesh(MeshIndex)
  For i = 1 To .Features.NumGroup
   If i <> CurrentFace Then
    For x = 0 To .Features.Groups(i).IndicesCount - 3
     'g = RayTri(Vert(.Features.Vertex(.Features.Groups(i).VertexOffset + x)), Vert(.Features.Vertex(.Features.Groups(i).VertexOffset + x + 1)), Vert(.Features.Vertex(.Features.Groups(i).VertexOffset + x + 2)), TDest2, TOrigin, t, 0, 0)
     g = Ray_Tri(TOrigin, TDest2, Vert(.Features.Vertex(.Features.Groups(i).VertexOffset + x)), Vert(.Features.Vertex(.Features.Groups(i).VertexOffset + x + 1)), Vert(.Features.Vertex(.Features.Groups(i).VertexOffset + x + 2)), t, 0, 0)
     
     If (g = 1 And t < DistMax - 1 And t >= 1) Then
          CheckFaces = 1
          Exit Function
     End If
   Next x
  End If
  Next i
 End With
 
End Function
Friend Function GetMeshIndex() As Integer
    GetMeshIndex = MeshIndex
End Function

Function GetPosition() As D3DVECTOR
   '##BD Returns the position of the mesh.
   GetPosition = Mesh(MeshIndex).Features.Settings.Position
End Function
Function GetRotation() As D3DVECTOR
   '##BD Returns the rotation of the mesh as 3 rotate angles in degrees.
   GetRotation = Mesh(MeshIndex).Features.Settings.Rotation
End Function

Public Sub LoadFromTVO(Filename As String)
 '##BD Load the mesh from a single TVO file. This method shoudn't be called
 '##PD Filename TVO file to load.
 
 'Load from the TrueVision format file.
 LoadMeshFromTVO Filename, MeshIndex, True
End Sub

Public Sub LoadMesh(Filename As String)
  '##BD Obsolete function. Shouldn't be called
  Open Filename For Binary As 1
    Dim a As String * 6
    Get #1, , a
    If a = "TVMESH" Then
      Get #1, , Mesh(MeshIndex).MeshName
  '    Get #1, , Mesh(MeshIndex).Features
    End If
  Close #1
End Sub

Public Sub LoadXFile(Filename As String)
   '##BD Load a simple non-animated X file with his materials and textures. The textures should be in the same directory than the model.
   '##PD Filename of the X file to load
     
    Dim Mat As D3DXBuffer, nbmat As Long, Mater() As D3DMATERIAL8
    
    Dim TexF As New TextureFactory8, temptex As Texture8
    Dim MatF As New MaterialFactory8, matind As Integer, tex As String
    TexF.SetLocal
    MatF.SetLocal
    Set Mesh(MeshIndex).D3DMesh = D3DX.LoadMeshFromX(Filename, 0, Device, Mesh(MeshIndex).Adj, Mat, nbmat)
    ReDim Mater(nbmat)
    With Mesh(MeshIndex).Features
    ReDim .Groups(nbmat)
    .NumGroup = nbmat
    Dim i As Long
    For i = 0 To nbmat - 1

        'copy material out of material buffer into our own structure
        D3DX.BufferGetMaterial Mat, i, Mater(i)
        
        Mater(i).Diffuse.a = 1
        Mater(i).emissive.a = 1
        Mater(i).Ambient.a = 1
'        Mater(i).Ambient = Mater(i).Ambient
         
        matind = MatF.CreateMaterialQuick(Mater(i).Diffuse.r, Mater(i).Diffuse.g, Mater(i).Diffuse.b, 1, "Xfilemat")
        'MatF.SetMat8 matind, Mater(i)
        .Groups(i + 1).MaterialI = matind
        .Groups(i + 1).IndicesCount = 3
        
        If D3DX.BufferGetTextureName(Mat, i) <> "" Then
          tex = D3DX.BufferGetTextureName(Mat, i)
          If Exist(tex) = False Then tex = Left$(Filename, InStrRev(Filename, "\")) + tex
          Set temptex = TexF.LoadTexture(tex, tex)
          .Groups(i + 1).TextureI = temptex.GetTexIndex
        End If
    
    Next i
    .Settings.VertexBufferEnabled = True
    .NbVertex = Mesh(MeshIndex).D3DMesh.GetNumVertices
    End With

    Debug.Print NBMaterial
    Set TexF = Nothing
    Set MatF = Nothing
    Debug.Print NBMaterial
End Sub
Sub MoveMesh(Optional Speed As Single = 1, Optional CheckCollision As Boolean = False)
    '##BD Move the mesh according to the SetVelocity values and the gravity.
    '##PD Speed Speed of the mesh.
    '##PD CheckCollision Checks collision before moving the mesh.
    With Mesh(MeshIndex).Features.Settings
     Dim Pos2 As D3DVECTOR
      Pos2 = VAdd(VScale(.Velocity, Speed * SystemFrame / 1000), .Position)
      .Velocity.y = .Velocity.y - .Gravity * SystemFrame / 1000
      If CheckCollision = True Then
       If LocalScene.Collide(.Position, Pos2, TV_TESTTYPE_ACCURATETESTING) = True Then
        .Velocity = Vector(0, 0, 0)
       Else
        .Position = Pos2
       End If
      Else
       .Position = Pos2
      End If
    End With
End Sub

Sub MoveVertex(Vertex As Integer, x As Single, y As Single, z As Single, Optional relative As Boolean = False)
  '##BD DYNAMICAL MESHES : Move a vertex to another position. You need to know the vertex index to do this (it's returned by the AddVertex method)
  '##PD Vertex Vertex ID in the mesh
  '##PD x New position x value
  '##PD y New position y value
  '##PD z New position z value
  '##PD relative Decides if the new positions are relative to the older ones.
   MoveVertexInFace Vertex, 0, x, y, z, relative
End Sub

Sub MoveVertexInFace(Vertex As Integer, facegroup As Integer, x As Single, y As Single, z As Single, Optional relative As Boolean = False)
   
  '##BD DYNAMICAL MESHES : Move a vertex to another position. You need to know the vertex index and the face index to do this (it's returned by the AddVertex method)
  '##PD Vertex Vertex ID in the face group
  '##PD facegroup Group ID in the mesh
  '##PD x New position x value
  '##PD y New position y value
  '##PD z New position z value
  '##PD relative Decides if the new positions are relative to the older ones.
   With Mesh(MeshIndex)
     Dim V As Integer
     
     V = .Features.Groups(facegroup).VertexOffset + Vertex
     
     If relative = True Then
     .Features.Vertex(V).x = .Features.Vertex(V).x + x
     .Features.Vertex(V).y = .Features.Vertex(V).y + y
     .Features.Vertex(V).z = .Features.Vertex(V).z + z
     .Features.LVertex(V).x = .Features.LVertex(V).x + x
     .Features.LVertex(V).y = .Features.LVertex(V).y + y
     .Features.LVertex(V).z = .Features.LVertex(V).z + z
     Else
     .Features.Vertex(V).x = x
     .Features.Vertex(V).y = y
     .Features.Vertex(V).z = z
     .Features.LVertex(V).x = x
     .Features.LVertex(V).y = y
     .Features.LVertex(V).z = z
     End If
   End With
   
End Sub
Sub Optimize()
 '##BD Calls this function when your static mesh is completed. When the mesh is optimized, the rendering is faster but you can't modify it ! Don't optimize very small object like billboard or objects of less than 24 vertices.
 Dim TempMesh As D3DXMesh
 Dim i As Long, d As Long
 Dim NumFaces As Long
 Dim Table() As Long
 Dim Index() As Integer
 Dim NV As Long, NT As Integer
 Dim NumVertex As Integer
 NumFaces = 0
 NV = 0
 'Copy everything in a IndexBuffer & a VertexBuffer
 'for optimization
 AddLog "MESH CLASS : Optimize the mesh "
 With Mesh(MeshIndex).Features
  For i = 1 To .NumGroup
    
    If .PrimType = TV_TRIANGLESTRIP Then
      NumFaces = NumFaces + .Groups(i).IndicesCount - 2
      ReDim Preserve Table(NumFaces)
      ReDim Preserve Index(NumFaces * 3)
      For d = 0 To .Groups(i).IndicesCount - 3
          Index(NV) = d + .Groups(i).VertexOffset
          Index(NV + 1) = d + .Groups(i).VertexOffset + 1
          Index(NV + 2) = d + .Groups(i).VertexOffset + 2
          NV = NV + 3
          
          Table(NT) = i - 1
          NT = NT + 1
      Next d
    Else
      
      NumFaces = NumFaces + Int(.Groups(i).IndicesCount / 3)
      ReDim Preserve Table(NumFaces)
      ReDim Preserve Index(NumFaces * 3)
      For d = 0 To .Groups(i).IndicesCount - 1
          Index(NV) = d + .Groups(i).VertexOffset
          NV = NV + 1
      Next d
      For d = 0 To (.Groups(i).IndicesCount - 1) \ 3
          Table(NT) = i - 1
          NT = NT + 1
      Next d

    End If
      
  Next i
  NumVertex = .NbVertex
   
  Dim V1 As D3DVERTEX2
  Dim V2 As D3DLVERTEX2
  Set Mesh(MeshIndex).D3DMesh = Nothing
  If .TypeVertex = TV_COLOR_VERTEX Then
    Set Mesh(MeshIndex).D3DMesh = D3DX.CreateMeshFVF(NumFaces, 1# * NumVertex, 0, TV_LVERTEX2, Device)
    D3DXMeshVertexBuffer8SetData Mesh(MeshIndex).D3DMesh, 0, 1# * NumVertex * Len(V2), 0, .LVertex(0)
  Else
    Set Mesh(MeshIndex).D3DMesh = D3DX.CreateMeshFVF(NumFaces, 1# * NumVertex, 0, TV_VERTEX2, Device)
    D3DXMeshVertexBuffer8SetData Mesh(MeshIndex).D3DMesh, 0, 1# * NumVertex * Len(V1), 0, .Vertex(0)
  End If
  D3DXMeshIndexBuffer8SetData Mesh(MeshIndex).D3DMesh, 0, 1# * NV * 2, 0, Index(0)
  
  
  Dim Pointer As Long
  Pointer = Mesh(MeshIndex).D3DMesh.LockAttributeBuffer(0)
  DXCopyMemory ByVal Pointer, Table(0), 4# * NumFaces
  Mesh(MeshIndex).D3DMesh.UnlockAttributeBuffer
  
  'Set Mesh(MeshIndex).D3DMesh = TempMesh
  'Set TempMesh = Nothing
  
End With
End Sub
Sub Render()
  '##BD Renders the mesh to the current scene. Useful if you want to render selected meshs and not all meshes.
'  RenderMesh MeshIndex, False
  RenderMesh2 MeshIndex, False

End Sub
Sub RenderGroupRange(StartGroup As Long, EndGroup As Long)
   '##BD Render partially a mesh.
   '##PD StartGroup From this Group ID
   '##PD EndGroup To this Group ID
   RenderMesh MeshIndex, True, StartGroup, EndGroup
End Sub



Sub SaveLightmaps(Filename As String)
  '##BD Save the lightmaps in a file to be reloaded later.
  '##PD Filename Filename of the lightmap file
  Dim Surface As Direct3DSurface8, Surf2 As Direct3DSurface8
  Dim Data() As Byte
  Dim w As Long, h As Long
  Dim Desc As D3DSURFACE_DESC
  Dim i As Long
   Open Filename For Binary As 1
  
  For i = 1 To Mesh(MeshIndex).Features.NumGroup
   If Mesh(MeshIndex).Features.Groups(i).LightmapI > 0 Then
     Set Surface = MeshLM(Mesh(MeshIndex).Features.Groups(i).LightmapI).GetSurfaceLevel(0)
     Surface.GetDesc Desc
      
     Set Surf2 = Device.CreateImageSurface(Desc.width, Desc.height, Desc.Format)
     
     Device.CopyRects Surface, ByVal 0, 0, Surf2, ByVal 0
     Surf2.GetDesc Desc
     
     ReDim Preserve Data(Desc.Size) As Byte
     Dim RECT As D3DLOCKED_RECT
     Surf2.LockRect RECT, ByVal 0, 0
     DXCopyMemory Data(0), ByVal RECT.pBits, Desc.Size
     Surf2.UnlockRect
     
     Put #1, , Desc.width
     Put #1, , Desc.height
     Put #1, , Desc.Format
     Put #1, , Desc.Size
     Put #1, , Data()
   End If
  Next i
 Close #1

End Sub

Public Sub SaveMesh(Filename As String)
  '##BD This function shouldn't be used. It's obsolete
  Open Filename For Binary As 1
     Put #1, , "TVMESH"
     Put #1, , Mesh(MeshIndex).MeshName
     Put #1, , Mesh(MeshIndex).Features
  Close #1
End Sub
Public Sub SaveToTVO(Filename As String, Optional TexDesc As Boolean = False)
   '##BD Save this mesh to a single TVO
   '##PD Filename TVO filename
   '##PD TexDesc Writes the texture table too
   SaveMeshToTVO Filename, MeshIndex, TexDesc
End Sub


Sub ScaleMesh(x As Single, y As Single, z As Single)
     '##BD Scales a mesh. <1 reduces the size. >1 improves the size. The x,y,z values should be the same in order to keep the volume ratio.
     '##PD x X-scale value
     '##PD y Y-scale value
     '##PD z Z-scale value
     Mesh(MeshIndex).Features.Settings.Scaling = Vector(x, y, z)
     Mesh(MeshIndex).Changed = True

End Sub

Sub SetBlendingMode(Optional Blending As CONST_TV_BLENDINGMODE = TV_BLEND_NO, Optional Groups As Integer = -1)
 '##BD Change the blending mode of a mesh/face object
 '##BD Blending Blending type.
 '##PD Groups Don't put anything to apply this in the entire mesh, else put the Group ID.
 Dim i As Long
 For i = 1 To Mesh(MeshIndex).Features.NumGroup
   If Groups = i Or Groups = -1 Then
     Mesh(MeshIndex).Features.Groups(i).AlphaBlending = Blending
   End If
 Next i
End Sub

Sub SetColor(color As Long, Optional Vertex As Integer = -1)
  '##BD Change the mesh color for a COLOR_VERTEX mesh.
  '##PD color New color, use RGBA macro to return the color from alpha, red, green and blue values.
  '##PD Vertex Put nothing to apply the color to the entire mesh, else put the Vertex id.
  Dim i As Long
  For i = 0 To Mesh(MeshIndex).Features.NbVertex
     If i = Vertex Or Vertex = -1 Then
       Mesh(MeshIndex).Features.LVertex(i).color = color
     End If
  Next i
End Sub

Public Sub SetCullMode(Cull As CONST_TV_CULLING)
   '##BD Change the BackFace culling for the mesh.
   '##PD Cull Culling mode. DoubleSided and Front are most used.
   Mesh(MeshIndex).CullMode = Cull
End Sub

Sub SetDeathTime(milliseconds As Single)
   '##BD Set the death time for the mesh. When the mesh time counter will reach this time, the mesh will be disabled
   '##PD milliseconds Death delay in milliseconds.
   Mesh(MeshIndex).DeathTime = milliseconds
End Sub

Public Sub SetGroupPointer(Group As Integer)
    '##BD Set the group ID pointer to the mesh. For advanced users.
    '##PD Group Group ID
    GroupPointer = Group
    If GroupPointer >= Mesh(MeshIndex).Features.NumGroup - 1 Then
       ReDim Preserve Mesh(MeshIndex).Features.Groups(Group)
       Mesh(MeshIndex).Features.NumGroup = Group
       Mesh(MeshIndex).Features.Groups(Group).IndicesCount = 0
       Mesh(MeshIndex).Features.Groups(Group).VertexOffset = Mesh(MeshIndex).Features.NbVertex
    End If
End Sub


Sub SetLODIndex(LODIndex As Single)
 '##BD Set the LOD for the mesh. You shound't call this function each frame but only if useful (5% slower)
 '##PD LODIndex 1: very accurate (original mesh) 5:not accurate
    If Mesh(MeshIndex).Features.Settings.ProgressiveMesh = True And LODIndex > 0 Then
    Mesh(MeshIndex).ProgMesh.SetNumVertices Mesh(MeshIndex).ProgMesh.GetMaxVertices / LODIndex
  End If
End Sub

Sub SetLODVertexes(NumVertices As Long)
 '##BD Set the wanted vertexes number for this mesh. Note that the LOD can alter considerably the mesh but can really speed the rendering of complex meshes. You shound't call this function each frame but only if useful (5% slower)
 '##PD NumVertices Wanted vertexes number (Between 1 and the Max number of vertexes)
  If Mesh(MeshIndex).Features.Settings.ProgressiveMesh = True Then
    Mesh(MeshIndex).ProgMesh.SetNumVertices NumVertices
  End If
End Sub

Public Sub SetMatrix(Matrix As D3DMATRIX)
  '##BD Set the transformation matrix for this mesh. for Advanced users.
  Mesh(MeshIndex).Matrix = Matrix
  Mesh(MeshIndex).Changed = False
End Sub

Friend Sub SetMeshIndex(SMeshIndex As Integer)
   MeshIndex = SMeshIndex
End Sub


Sub SetMeshName(MeshName As String)
  '##BD Changes the current name of the mesh. Remember you can use the macro to get a mesh : GetMesh("MeshName")
  '##PD MeshNew New name for the mesh
  Mesh(MeshIndex).MeshName = MeshName
End Sub

Sub SetPosition(x As Single, y As Single, z As Single)
   '##BD Changes the position of the mesh.
   '##PD x X-value position
   '##PD y Y-value position
   '##PD z Z-value position
   Mesh(MeshIndex).Features.Settings.Position = Vector(x, y, z)
   Mesh(MeshIndex).Changed = True
End Sub



Public Sub SetPrimitiveType(Primitive As CONST_TV_PRIMITIVETYPE)
    '##BD Changes the primitive type of the mesh. For advanced users only
    Mesh(MeshIndex).Features.PrimType = Primitive

End Sub

Sub SetRotation(x As Single, y As Single, z As Single)
  '##BD Set the rotation of the mesh around the x,y,z axis. IN DEGREES
  '##PD x Rotation around X-axis IN DEGREES
  '##PD y Rotation around Y-axis IN DEGREES
  '##PD z Rotation around Z-axis IN DEGREES
  Mesh(MeshIndex).Features.Settings.Rotation = Vector(x, y, z)
     Mesh(MeshIndex).Changed = True

End Sub

Sub SetShader(Shader As Shader8, Optional Group As Integer = -1)
    
  '##BD Set the shader for the entire mesh or for a specifical mesh group. See the shader tutorial for further information.
  '##PD Shader Shader to apply to the mesh
  '##PD Group Put nothing to apply the shader to the entire mesh, else put the group ID
  
  
    On Error Resume Next
    Dim ShInd As Long, i As Long
    
    If Shader Is Nothing Then
       ShInd = 0
    Else
       ShInd = Shader.GetShaderIndex
    End If
    With Mesh(MeshIndex)
    For i = 1 To .Features.NumGroup
       If Group = i Or Group = -1 Then
          .Features.Groups(i).ShaderI = ShInd
       End If
    Next i
    End With
End Sub

Sub SetTexture(Texture As Texture8, Optional Group As Single = -1)
    '##BD Set the texture for the entire mesh or for a specifical mesh group.
    '##PD Texture Texture object to be a applied. You can use the GetTex("TexName") macro.
    '##PD Group Put nothing if the texture have to be applied to the entire mesh, else put the group ID
    
    On Error Resume Next
    Dim TxInd As Long, i As Long
    
    If Texture Is Nothing Then
       TxInd = 0
    Else
       TxInd = Texture.GetTexIndex
    End If
    With Mesh(MeshIndex)
    For i = 1 To .Features.NumGroup
       If Group = i Or Group = -1 Then
          .Features.Groups(i).TextureI = TxInd
       End If
    Next i
    End With
End Sub

Sub SetMaterial(MaterialIndex As Integer, Optional Group As Single = -1)
    '##BD Set the material created by the MatFactory for the mesh or a group ID.
    '##PD MaterialIndex Index returned my the MatFactory creation function. Can be 0 for white material.
    '##PD Group Put nothing to apply the material to the entire mesh, else put the Group ID.
    'Set the material for the entire mesh or for a specifical mesh group.
    Dim MaInd As Long, i As Long
    On Error Resume Next
    With Mesh(MeshIndex)
    For i = 1 To .Features.NumGroup
       If Group = i Or Group = -1 Then
          .Features.Groups(i).MaterialI = MaterialIndex
       End If
    Next i
    End With
End Sub
Sub SetVelocity(x As Single, y As Single, z As Single, Optional Gravity As Single = 0)
  '##BD Set the current velocity for the mesh.MoveMesh function.
  '##PD x velocity X value
  '##PD y velocity Y value
  '##PD z velocity Z value
  
  Mesh(MeshIndex).Features.Settings.Gravity = Gravity
  Mesh(MeshIndex).Features.Settings.Velocity = Vector3(x, y, z)
End Sub
Public Sub SetVertexType(TypeV As CONST_TV_VERTEXTYPE)
   '##BD Change the vertex type. Usually you should call this function before adding vertexes or loading files.
   '##PD TypeV COLOR_VERTEX will use the vertex color and won't be lit. The NORMAL_VERTEX can be lit.
    Mesh(MeshIndex).Features.TypeVertex = TypeV
End Sub

Private Sub Class_Terminate()
  Set Mesh(MeshIndex).Adj = Nothing
  Set Mesh(MeshIndex).D3DMesh = Nothing
  Set Mesh(MeshIndex).ProgMesh = Nothing
  Set Mesh(MeshIndex).D3DMeshSkin = Nothing
  Mesh(MeshIndex).MeshName = ""
  Mesh(MeshIndex).Active = False
  Mesh(MeshIndex).Enabled = False
  ReDim Mesh(MeshIndex).Features.Groups(0)
  Mesh(MeshIndex).CullMode = D3DCULL_NONE
  Mesh(MeshIndex).Features.NumGroup = 0
End Sub

