Attribute VB_Name = "TrueVisionGlobal"
'--------------------------------------'
' TrueVision 8  -  Global shared module
'--------------------------------------'

' General Declarations

Public DirectX As New DirectX8
Public D3DX As New D3DX8

' Direct3D
Public D3D As Direct3D8
Public Device As Direct3DDevice8

' DirectInput
Public DInput As DirectInput8
Public DIKeyboardDevice As DirectInputDevice8
Public DIMouseDevice As DirectInputDevice8
Public lKeyEvent As Long
Public lMouseEvent As Long

 'States
Public KeyboardState As DIKEYBOARDSTATE
Public MouseState As DIMOUSESTATE

' App 3D Global Type
Public Type TrueVisionApp
  Fullscreen As Boolean
  PresentMode As D3DPRESENT_PARAMETERS
  hwnd As Long
  DeviceType As Long
  Initialized As Boolean
  KeyboardInit As Boolean
  Caps As D3DCAPS8
  ParticleSize As Single
End Type

Public TVApp As TrueVisionApp


'Maths, rasterizer and camera settings
'Pi
Global Const PI = 3.141592653

'Matrices
Public ViewMatrix As D3DMATRIX, ProjectionMatrix As D3DMATRIX
Public WorldMatrix As D3DMATRIX
'Camera
Public StartPos As D3DVECTOR, EndPos As D3DVECTOR, CameraRoll As Single
Public AngleFOV As Single, farPlane As Single, Aspect As Single
'Viewport
Public Viewport As D3DVIEWPORT8, BackGroundColor As Long

'Global FVF Vertex format
Global Const TV_VERTEX = (D3DFVF_XYZ Or D3DFVF_NORMAL Or D3DFVF_TEX1)
Global Const TV_VERTEX2 = (D3DFVF_XYZ Or D3DFVF_NORMAL Or D3DFVF_TEX2)
Global Const TV_LVERTEX = (D3DFVF_XYZ Or D3DFVF_DIFFUSE Or D3DFVF_SPECULAR Or D3DFVF_TEX1)
'Global Const TV_LVERTEX2 = (D3DFVF_XYZ Or D3DFVF_DIFFUSE Or D3DFVF_SPECULAR Or D3DFVF_TEX2) 'Or D3DFVF_TEXCOORDSIZE2(0) Or D3DFVF_TEXCOORDSIZE2(1))
Global Const TV_LVERTEX2 = 706
Global Const TV_TLVERTEX = (D3DFVF_XYZRHW Or D3DFVF_DIFFUSE Or D3DFVF_SPECULAR Or D3DFVF_TEX1)

 Public MouseX As Long, MouseY As Long, ActMX As Long, ActMY As Long, RelMX As Long, RelMY As Long, AmouseX As Long, AmouseY As Long
 Public relative As Boolean

'Global time functions
Public SystemFrame As Long, FPS As Single, OldFPS As Long, OldSys As Long
Public NBImage As Long, MultiTexturing As Boolean
Public FPSDraw As Boolean

'Global API function
Public Declare Function GetTickCount Lib "kernel32" () As Long

'Fonts Func
Public D3DFont As D3DXFont

Public MeshLM() As Direct3DTexture8, NumMeshLM As Integer

' Global Landscape & BSP Tree
Public LocalScene As New Scene8
Public BSPTree As BSPTree8
Public LocalInput As InputEngine8
Public LocalLand As Landscape8
Public LocalScreen As New Screen8
Public LocTexFactory As New TextureFactory8
'-------------------------
' PAK files Declarations
'------------------------------

Public Type FileInPAK
   named As String * 56
   filepos As Long
   filelen As Long
End Type
 
Public Type PAKFileHeader
  Ident As Long
  dirofs As Long
  dirlen As Long
End Type

Const MAX_FILES_IN_PACK = 4096


Public PakHeader As PAKFileHeader
Public PakFiles(MAX_FILES_IN_PACK) As FileInPAK
Public NumPakFiles As Integer
Public PakPath As String * 64
Public FileToDelete As Boolean
Public PackedFile As String
Public FilePAK As String
Public FileLoaded() As String
Public PakEnable As Boolean, NBFileInPak

Public DisplayConsole As Boolean

Public ErrStr As String, PrecisionCollide As Single, PrecisionMousePicking As Single

Public Type BITMAPINFOHEADER  '40 bytes
        biSize As Long
        biWidth As Long
        biHeight As Long
        biPlanes As Integer
        biBitCount As Integer
        biCompression As Long
        biSizeImage As Long
        biXPelsPerMeter As Long
        biYPelsPerMeter As Long
        biClrUsed As Long
        biClrImportant As Long
End Type
Public Type BITMAPHEADER
     magic As Integer
     Size As Long
     Reserved As Integer
     Reserved2 As Integer
     offset As Long
End Type
Public Type BITMAPHEADER2
     magic As Integer
     Size1 As Integer
     Size2 As Integer
     Reserved As Integer
     Reserved2 As Integer
     offset As Integer
     offset2 As Integer
End Type

Public Declare Function GetClientRect Lib "user32" (ByVal hwnd As Long, lpRect As RECT) As Long
Public Declare Function GetPixel Lib "gdi32" (ByVal hdc As Long, ByVal x As Long, ByVal y As Long) As Long
Public Sub QSort(z() As Integer, Ix() As Integer, de As Integer, a As Integer)
  Dim i As Integer, j As Integer, x As Integer
  Dim t As Integer
  i = de: j = a
  x = z((i + j) / 2)
  
  Do
   While z(i) < x: i = i + 1: Wend
   While z(j) > x: j = j - 1: Wend
   If i <= j Then
      
      t = z(i)
      z(i) = z(j)
      z(j) = t
      
      t = Ix(i)
      Ix(i) = Ix(j)
      Ix(j) = t
      
      i = i + 1
      j = j - 1
      
   End If
  Loop Until i > j
  
  If j > de Then QSort z(), Ix(), de, j
  If i < a Then QSort z(), Ix(), i, a
End Sub

Public Function ClassifyPlane(P As D3DVECTOR, Plane As TVPlane) As Integer
' int ClassifyPoint(D3DXVECTOR3 *pos,PLANE * Plane)
'{
'float result;
'D3DXVECTOR3 *vec1=(D3DXVECTOR3 *)&Plane->PointOnPlane;
'D3DXVECTOR3 Direction=(*vec1)-(*pos);
'result=D3DXVec3Dot(&Direction,&Plane->Normal);
'if (result<-0.001) return CP_FRONT;
'if (result>0.001) return CP_BACK;'''

'return CP_ONPLANE;
'}

   Dim t As Single
   t = D3DXVec3Dot(P, Plane.normal) - Plane.offset
   If t < -0.001 Then ClassifyPlane = -1: Exit Function
   If t > 0.001 Then ClassifyPlane = 1: Exit Function
   ClassifyPlane = 0


End Function

Sub ComputeClipPlanes()

Dim ViewProj As D3DMATRIX
D3DXMatrixMultiply ViewProj, ViewMatrix, ProjectionMatrix


With FrustumPlanes(0)
  .a = (ViewProj.m14 + ViewProj.m11)
  .b = (ViewProj.m24 + ViewProj.m21)
  .C = (ViewProj.m34 + ViewProj.m31)
  .d = (ViewProj.m44 + ViewProj.m41)
End With
  
With FrustumPlanes(1)
  .a = (ViewProj.m14 - ViewProj.m11)
  .b = (ViewProj.m24 - ViewProj.m21)
  .C = (ViewProj.m34 - ViewProj.m31)
  .d = (ViewProj.m44 - ViewProj.m41)
End With
  
With FrustumPlanes(2)
  .a = (ViewProj.m14 - ViewProj.m12)
  .b = (ViewProj.m24 - ViewProj.m22)
  .C = (ViewProj.m34 - ViewProj.m32)
  .d = (ViewProj.m44 - ViewProj.m42)
End With
  
With FrustumPlanes(3)
  .a = (ViewProj.m14 + ViewProj.m12)
  .b = (ViewProj.m24 + ViewProj.m22)
  .C = (ViewProj.m34 + ViewProj.m32)
  .d = (ViewProj.m44 + ViewProj.m42)
End With
  
With FrustumPlanes(4)
  .a = (ViewProj.m14 + ViewProj.m13)
  .b = (ViewProj.m24 + ViewProj.m23)
  .C = (ViewProj.m34 + ViewProj.m33)
  .d = (ViewProj.m44 + ViewProj.m43)
End With
  
With FrustumPlanes(5)
  .a = (ViewProj.m14 - ViewProj.m13)
  .b = (ViewProj.m24 - ViewProj.m23)
  .C = (ViewProj.m34 - ViewProj.m33)
  .d = (ViewProj.m44 - ViewProj.m43)
End With
  
Dim Denom As Single, i As Long
For i = 0 To 5
 With FrustumPlanes(i)
  D3DXPlaneNormalize FrustumPlanes(i), FrustumPlanes(i)
 End With
Next i


End Sub

'Equivalent to D3DFVF_TEXCOORDSIZE1_coordIndex, specifying that one
'texture coordinate is used for the vertex format specified by coordIndex
Function D3DFVF_TEXCOORDSIZE1(coordIndex As Long) As Long
    D3DFVF_TEXCOORDSIZE1 = D3DFVF_TEXTUREFORMAT1 * 2 ^ ((coordIndex * 2) + 16)
End Function

'Equivalent to D3DFVF_TEXCOORDSIZE1_coordIndex, specifying that two
'texture coordinates are used for the vertex format specified by coordIndex
Function D3DFVF_TEXCOORDSIZE2(coordIndex As Long) As Long
    D3DFVF_TEXCOORDSIZE2 = D3DFVF_TEXTUREFORMAT2
End Function

'Equivalent to D3DFVF_TEXCOORDSIZE1_coordIndex, specifying that three
'texture coordinates are used for the vertex format specified by coordIndex
Function D3DFVF_TEXCOORDSIZE3(coordIndex As Long) As Long
    D3DFVF_TEXCOORDSIZE3 = D3DFVF_TEXTUREFORMAT3 * 2 ^ ((coordIndex * 2) + 16)
End Function

'Equivalent to D3DFVF_TEXCOORDSIZE1_coordIndex, specifying that four
'texture coordinate are used for the vertex format specified by coordIndex
Function D3DFVF_TEXCOORDSIZE4(coordIndex As Long) As Long
    D3DFVF_TEXCOORDSIZE4 = D3DFVF_TEXTUREFORMAT4 * 2 ^ ((coordIndex * 2) + 16)
End Function

 
Public Function ClassifyPoint2(P As D3DVECTOR, Plane As TVPlane)
   Dim t As Single
   t = D3DXVec3Dot(P, Plane.normal)
   If t < Plane.offset Then ClassifyPoint = -1 Else ClassifyPoint = 1
End Function

Sub TexSave(file As String, tex As Direct3DTexture8)
  'Take a screen shot of the current texture
    
  Dim Surface As Direct3DSurface8, Surf2 As Direct3DSurface8
  
  Dim Data() As Byte
  Set Surface = tex.GetSurfaceLevel(0)
  
  Dim w As Long, h As Long
  Dim Desc As D3DSURFACE_DESC
  Surface.GetDesc Desc
   
  Debug.Print Desc.width, Desc.Format, Desc.height, Desc.Size
  Set Surf2 = Device.CreateImageSurface(Desc.width, Desc.height, Desc.Format)
  
  Device.CopyRects Surface, ByVal 0, 0, Surf2, ByVal 0
  Surf2.GetDesc Desc
  
  ReDim Preserve Data(Desc.Size) As Byte
  Dim RECT As D3DLOCKED_RECT
  'If TVApp.Fullscreen = True Then
   
   Surf2.LockRect RECT, ByVal 0, 0
   DXCopyMemory Data(0), ByVal RECT.pBits, Desc.Size
   Surf2.UnlockRect
   
    Dim MemoryCount As Long, y As Long, temp As Byte
   Dim data2() As Byte
   ReDim data2(Desc.Size) As Byte
   For y = 0 To Desc.height - 1
      DXCopyMemory data2(y * RECT.Pitch), Data((Desc.height - y - 1) * RECT.Pitch), RECT.Pitch
   Next y
    Dim offset As Long, offset2 As Long
    offset = 0
    offset2 = 0
   
   If Desc.Format = D3DFMT_R5G6B5 Then
    '16bits modes
    ReDim Data3(Desc.Size * 3 / 2) As Byte
     For i = 0 To Desc.Size / 2 - 1
      Data3(offset + 2) = data2(offset2 + 1) And &HF8
      Data3(offset + 1) = (data2(offset2 + 1) And &H7) * 32 + (data2(offset2) And &HE0) \ 8
      Data3(offset + 0) = (data2(offset2) And &H1F) * 8
      offset = offset + 3
      offset2 = offset2 + 2
     Next i
   Else
    '32bits modes, alpha is unsignifiant
    ReDim Data3(Desc.Size * 3 / 4) As Byte
     For i = 0 To Desc.Size / 2 - 1
      Data3(offset + 2) = data2(offset2 + 1)
      Data3(offset + 1) = data2(offset2 + 2)
      Data3(offset + 0) = data2(offset2 + 3)
      offset = offset + 3
      offset2 = offset2 + 4
     Next i
   End If
  MemoryCount = UBound(Data3) + 54
  Dim bm As BITMAPHEADER
  Dim bm2 As BITMAPINFOHEADER
  bm.magic = 19778
  bm.Size = MemoryCount
  bm.offset = 54
  bm2.biBitCount = 24
  bm2.biWidth = Desc.width
  bm2.biHeight = Desc.height
  bm2.biSize = 40
  bm2.biPlanes = 1
  bm2.biSizeImage = MemoryCount - 54
  bm2.biCompression = 0
  bm2.biXPelsPerMeter = 50
  bm2.biYPelsPerMeter = 50
   
   Close 1
   Open file For Binary As 1
    Put #1, , bm
    Put #1, , bm2
    Put #1, , Data3()
       
   Close
  'End If
  Erase Data()
  Set Surface = Nothing
End Sub

Function CheckBoxVisible(BoxMin As D3DVECTOR, BoxMax As D3DVECTOR) As Integer
    Dim TMin As D3DVECTOR
    Dim TMax As D3DVECTOR
    Dim Matrix As D3DMATRIX
    Dim i As Long, P As Long, Dist As Single, clipped As Integer
    
    Device.GetTransform D3DTS_WORLD, Matrix
    
    TMin = MTransform(Matrix, BoxMin)
    TMax = MTransform(Matrix, BoxMax)
    
    Dim TestV(9) As D3DVECTOR
    
    TestV(1) = TMax
    TestV(2) = Vector(TMax.x, TMax.y, TMin.z)
    TestV(3) = Vector(TMax.x, TMin.y, TMax.z)
    TestV(4) = Vector(TMax.x, TMin.y, TMin.z)
    TestV(5) = Vector(TMin.x, TMax.y, TMax.z)
    TestV(6) = Vector(TMin.x, TMax.y, TMin.z)
    TestV(7) = Vector(TMin.x, TMin.y, TMax.z)
    TestV(8) = Vector(TMin.x, TMin.y, TMin.z)
    TestV(9) = Vector((TMin.x + TMax.x) / 2, (TMin.y + TMax.y) / 2, (TMin.z + TMax.z) / 2)
    
    For i = 1 To 9
       clipped = 0
       For P = 0 To 5
          
          Dist = D3DXPlaneDotCoord(FrustumPlanes(P), TestV(i))
         ' Debug.Print i, p, dist
          
          
          If Dist < 0 Then
             clipped = 1
             Exit For
          End If
       Next P
       If clipped = 0 Then
         CheckBoxVisible = 1
         Exit Function
       End If
    Next i
    'Debug.Print "No visible"
    CheckBoxVisible = 0
End Function
Function CheckSphere(Center As D3DVECTOR, Radius As Single) As Integer
   Dim TCenter As D3DVECTOR
   Dim Matrix As D3DMATRIX, i As Long
   On Error Resume Next
    
   Device.GetTransform D3DTS_WORLD, Matrix
   
   TCenter = MTransform(Matrix, Center)
   
   For i = 0 To 5
        Dist = D3DXPlaneDotCoord(FrustumPlanes(i), TCenter)
        If Dist < -Radius - 30 Then
           CheckSphere = 0
           Exit Function
        End If
   Next i
   CheckSphere = 1
   
End Function

'-----------------------------------------------------------------------------
' Name: D3DUtil_ComputeClipPlanes
'-----------------------------------------------------------------------------
'
'
Sub ComputeClipPlanes2()
    
    Dim veye As D3DVECTOR
    Dim vat As D3DVECTOR
    Dim vup As D3DVECTOR
    Dim Front As Single
    Dim Back As Single
    Dim Fov As Single
    
    veye = StartPos
    vat = EndPos
    vup = Vector(0, 1, 0)
    Front = 1
    Back = farPlane
    Fov = AngleFOV
    'Aspect = 3 / 4
    
    Dim vDir As D3DVECTOR
    Dim vright As D3DVECTOR
        
    Dim vFrontCenter As D3DVECTOR
    Dim vFrontUp As D3DVECTOR
    Dim vFrontRight As D3DVECTOR
    
    Dim vBackCenter As D3DVECTOR
    
    Dim vBackRight As D3DVECTOR
    Dim vbackLeft As D3DVECTOR
    
    Dim vBackRightTop As D3DVECTOR
    Dim vBackLeftTop As D3DVECTOR
    
    Dim vBackRightBot As D3DVECTOR
    Dim vBackLeftBot As D3DVECTOR
        
    Dim dx As Single
    Dim dy As Single
    
    'Establish our basis vector
    D3DXVec3Subtract vDir, vat, veye
    D3DXVec3Normalize vDir, vDir
    D3DXVec3Normalize vup, vup
    D3DXVec3Cross vright, vDir, vup
    
    dx = Tan(Fov / 2) * Back
    dy = dx * Aspect
        
    '
    '
    '              /|  vbackleft (top,bot)
    '             / |
    '        vfront |
    '           /|  |
    '       eye ----|  vbackcenter
    '           \|  |
    '            \  |dx
    '             \ |
    '              \|  vbackright (top,bot)
    '
    '
    
    
    'compute vbackcenter
    D3DXVec3Scale vBackCenter, vDir, Back
    D3DXVec3Add vBackCenter, vBackCenter, veye
    
    'Debug.Print dx, dy
    'compute vbackright
    D3DXVec3Scale vBackRight, vright, dx * 1.2 '- farPlane * 0.5
    D3DXVec3Add vBackRight, vBackCenter, vBackRight
    
    
    'compute vbackleft
    D3DXVec3Scale vbackLeft, vright, -dx '- farPlane * 0.5
    D3DXVec3Add vbackLeft, vBackCenter, vbackLeft
 
    'compute vbackrighttop
    D3DXVec3Scale vBackRightTop, vup, dy
    D3DXVec3Add vBackRightTop, vBackRight, vBackRightTop
    
    
    'compute vbacklefttop
    D3DXVec3Scale vBackLeftTop, vup, dy
    D3DXVec3Add vBackLeftTop, vBackRight, vBackLeftTop
        
     'compute vbackrightbot
    D3DXVec3Scale vBackRightBot, vup, -dy
    D3DXVec3Add vBackRightBot, vBackRight, vBackRightBot
       
    'compute vbackleftbot
    D3DXVec3Scale vBackLeftBot, vup, -dy
    D3DXVec3Add vBackLeftBot, vBackRight, vBackLeftBot
    
        
    'compute vfrontcenter
    D3DXVec3Scale vFrontCenter, vDir, Front
    D3DXVec3Add vFrontCenter, vFrontCenter, veye
 
    'compute vfrontright
    D3DXVec3Scale vFrontRight, vright, dx
    D3DXVec3Add vFrontRight, vFrontCenter, vFrontRight
 
    'compute vfrontup
    D3DXVec3Scale vFrontUp, vup, dy
    D3DXVec3Add vFrontUp, vFrontCenter, vFrontUp
 
    '
    'g_numClipPlanes = 6
    
    
    
    'front plane
    D3DXPlaneFromPointNormal FrustumPlanes(0), veye, vDir
    
    'back plane
    Dim vnegdir As D3DVECTOR
    D3DXVec3Scale vnegdir, vDir, -1
    D3DXPlaneFromPointNormal FrustumPlanes(1), vBackCenter, vnegdir
    
    'right plane
    D3DXPlaneFromPoints FrustumPlanes(2), veye, vBackRightTop, vBackRightBot
    
    'left plane
    D3DXPlaneFromPoints FrustumPlanes(3), veye, vBackLeftTop, vBackLeftBot
    
    'top plane
    D3DXPlaneFromPoints FrustumPlanes(4), veye, vBackLeftTop, vBackRightTop
    
    'bot plane
    D3DXPlaneFromPoints FrustumPlanes(5), veye, vBackRightBot, vBackLeftBot
    'g_numClipPlanes = 4
    
End Sub
Sub ClearScreen()
   Device.Clear 0, ByVal 0, D3DCLEAR_TARGET Or D3DCLEAR_ZBUFFER, 0, 1, 0
End Sub

Sub DeletePAKTemp(file As String)
  If file <> "" And PakEnable = True And FileToDelete = True And FilePAK = file Then
   FileToDelete = False
   'AddLog "Kill " + file
   Kill file
  End If
End Sub
Sub FlipToScreen()
   Device.Present ByVal 0, ByVal 0, 0, ByVal 0
End Sub

Function GetTex(Name As String) As Texture8
  For i = 1 To NBTexture
     If UCase$(Name) = UCase$(Textures(i).TextureName) Then
       Set GetTex = Textures(i).TextureOb
       Exit Function
     End If
  Next i
  Set GetTex = Nothing
End Function

Function SearchExistPath(Filename As String, Optional eventualPath As String, Optional Recursive As Boolean) As String
     Dim PathStr As Long
     
     If Exist(Filename) = True Then
       SearchExistPath = Filename
       Exit Function
     End If
     If Exist(eventualPath + Filename) = True Then
       SearchExistPath = eventualPath + Filename
       Exit Function
     End If
     If Exist(eventualPath + "\" + Filename) = True Then
       SearchExistPath = eventualPath + "\" + Filename
       Exit Function
     End If
     
     'Test Quake3 TGA->JPG file
     If UCase$(Right$(Filename, 3)) = "TGA" Then
          befile = Filename
          Mid(Filename, Len(Filename) - 2) = "JPG"
          If Exist(Filename) Then
            SearchExistPath = Filename
            Exit Function
          End If
          If Exist(eventualPath + Filename) Then
            SearchExistPath = eventualPath + Filename
            Exit Function
          End If
          Filename = befile
      End If
     
     'remove the filename path
     PathStr = InStrRev(Filename, "\") + InStrRev(Filename, "/")
     If PathStr <> 0 And Recursive = False Then
        'we check with the new file without path
        SearchExistPath = SearchExistPath(mid(Filename, PathStr + 1), eventualPath, True)
        Exit Function
     End If
     
      
     'If Right(eventualPath, 1) <> "\" Or Right(eventualPath, 1) <> "/" Then
        
     
     'else returns the filename to test in PAK files
     SearchExistPath = Filename
End Function
Function Exist(Filename As String) As Boolean
   On Local Error GoTo NoFound
   Open Filename For Input As 16
   Close #16
   Exist = True
   Exit Function
NoFound:
   Err.Clear
   Exist = False
End Function
Sub TestPAKFile(ByRef Filename As String)
 Dim FileIndex As Integer
 Dim i As Integer
 On Error GoTo SendError
 'test if we have to extract the file from the PAK
 If PakEnable = True Then
   AddLog "Expand " + Filename + " from Pak file. STATE : OK"
   FileIndex = -1
   If Dir(Filename) = "" Then
     For i = 0 To NBFileInPak
      If UCase(Filename) = Trim(UCase(PakFiles(i).named)) Then FileIndex = i: Exit For
     Next i
     If FileIndex = -1 Then Filename = "": AddLog "No found in PAK. STATE : ERROR (Level 2/5) ": Exit Sub
   
     With PakFiles(FileIndex)
      Filename = Trim(.named)
      'Replace "/" by "\"
       For i = 1 To Len(Filename)
         If mid(Filename, i, 1) = "/" Then Mid(Filename, i, 1) = "\"
       Next i
    
       'Get the filename without path
       For i = Len(Filename) To 1 Step -1
         If mid(Filename, i, 1) = "\" Then Exit For
       Next i
       Filename = mid(Filename, i + 1)
    
       Open Filename For Binary As 4
       ReDim a(.filelen + 1) As Byte
       Get #80, .filepos + 1, a
       Put #4, , a
       Close #4
       AddLog Filename + " has been expanded from Pak File. STATE : OK"
       FileToDelete = True
       FilePAK = Filename
     End With
  End If
 End If
Exit Sub
 
SendError:
 AddLog "File in PAK " + Filename + " has occured a error. STATE : ERROR (Level 2/5) "
 AddLog "Error description :" + Err.Description + " # Number " + CStr(Err.Number)
 Err.Clear
 
End Sub
Function RayTri(ByRef v0 As D3DVECTOR, ByRef V1 As D3DVECTOR, ByRef V2 As D3DVECTOR, vDir As D3DVECTOR, vOrig As D3DVECTOR, t As Single, U As Single, V As Single, Optional CNormal As Boolean = False) As Integer

    Dim Edge1 As D3DVECTOR
    Dim Edge2 As D3DVECTOR
    Dim pvec As D3DVECTOR
    Dim tvec As D3DVECTOR
    Dim qvec As D3DVECTOR
    Dim det As Single
    Dim fInvDet As Single
    
    'find vectors for the two edges sharing vert0
    D3DXVec3Subtract Edge1, V1, v0
    D3DXVec3Subtract Edge2, V2, v0
    
    'begin calculating the determinant - also used to caclulate u parameter
    D3DXVec3Cross pvec, vDir, Edge2
    
    'if determinant is nearly zero, ray lies in plane of triangle
    det = D3DXVec3Dot(Edge1, pvec)
    If Abs(det < 0.0001) Then
        Exit Function
    End If
    
    'calculate distance from vert0 to ray origin
    D3DXVec3Subtract tvec, vOrig, v0

    'calculate u parameter and test bounds
    U = D3DXVec3Dot(tvec, pvec)
    If (U < 0 Or U > det) Then
        Exit Function
    End If
    
    'prepare to test v parameter
    D3DXVec3Cross qvec, tvec, Edge1
    
    'calculate v parameter and test bounds
    V = D3DXVec3Dot(vDir, qvec)
    If (V < 0 Or (U + V > det)) Then
        Exit Function
    End If
    
    'calculate t, scale parameters, ray intersects triangle
    t = D3DXVec3Dot(Edge2, qvec)
    fInvDet = 1 / det
    t = t * fInvDet
    U = U * fInvDet
    V = V * fInvDet
    If t = 0 Then Exit Function
    If CNormal = True Then
      D3DXVec3Cross CollisionNormal, VNormalize(Edge1), VNormalize(Edge2)
      CollisionImpact = VAdd(vOrig, VScale(vDir, t))
    End If
    'CollisionNormal = D3D
    RayTri = 1
    
End Function
'
'
'
Sub AddLog(Desc As String, Optional Starting As Boolean = False)
 ' Add info to the log file
 On Error Resume Next
  
 If Starting = False Then
  
  Open "c:\debug8.txt" For Append As 52
   
   Print #52, Date$ + " " + Time$ + " | "; Desc
  
  Close #52
 
 Else
  
  Open "c:\debug8.txt" For Output As 52
   Print #52, "###################################################"
   Print #52, "#           TrueVision 8 Log file                 #"
   Print #52, "###################################################"
   Print #52, ""
   Print #52, "    Engine started at " + Date$ + " " + Time$
   Print #52, ""
  Close #52
 
 End If
 
End Sub


Sub TestTextureFormats()
 
 'Check global textures
 Dim Format As Long
 Format = D3DFMT_R5G6B5
 D3DX.CheckTextureRequirements Device, 256, 256, 1, 0, Format, D3DPOOL_DEFAULT
 AddLog "TEXTURE REQUIREMENTS CHECKING : D3DFMT_R5G6B5 : " + IIf(Format = 23, "Yes", "No")
 FmtTexture = Format
 If Format <> 23 Then
   'The device doesn't support 16 bits textures.
   'Let's try with 32 bits ones.
    Format = D3DFMT_A8R8G8B8
    D3DX.CheckTextureRequirements Device, 256, 256, 1, 0, Format, D3DPOOL_DEFAULT
    AddLog "TEXTURE REQUIREMENTS CHECKING : D3DFMT_A8R8G8B8 : " + IIf(Format = D3DFMT_A8R8G8B8, "Yes", "No")
    FmtTexture = Format
    If Format <> D3DFMT_A8R8G8B8 Then
     AddLog "TEXTURE REQUIEREMENTS CHECKING : TrueVision may have problems because your 3d card doesn't support the 16bits or 32bits texture mode "
     AddLog "TEXTURE REQUIEREMENTS CHECKING : Format that will be used " + CStr(Format)
     FmtTexture = Format
    End If
 Else
    
 
 'Check colorkeying for 16bits textures
 Format = D3DFMT_A4R4G4B4
 D3DX.CheckTextureRequirements Device, 256, 256, 1, 0, Format, D3DPOOL_DEFAULT
 AddLog "TEXTURE REQUIREMENTS CHECKING : D3DFMT_A1R5G6B5 : " + IIf(Format = D3DFMT_A4R4G4B4, "Yes", "No")
 FmtAlphaTexture = Format
 
 End If

' Format = D3DFMT_R5G6B5
' D3DX.CheckTextureRequirements Device, 256, 256, 1, 0, Format
' AddLog "TEXTURE REQUIREMENTS CHECKING : D3DFMT_R5G6B5 : " + IIf(Format = 23, "Yes", "No")
 
End Sub

Sub UnloadAll()
   On Error Resume Next
   ReDim Materials(0)
   Dim i As Long
   NBMaterial = 0
   AddLog "MATERIAL FACTORY : Desallocate Materials Memory"
   
   For i = 1 To NBTexture
     With Textures(i)
       If Not .TextureOb Is Nothing Then Set .TextureOb = Nothing
       If Not .TexSurf Is Nothing Then Set .TexSurf = Nothing
       If Not .RTS Is Nothing Then Set .RTS = Nothing
     End With
   Next i
   
   AddLog "TEXTURE FACTORY : Desallocate Textures Memory"
   ReDim Textures(0)
   NBTexture = 0
   Err.Clear
End Sub


