Attribute VB_Name = "TrueVisionGlobalTypes"
'-------------------------------
' TrueVision Global Types
'-------------------------------
' Meshes / Lights / Sound


' Started 19/11/00

Option Explicit

Public Declare Function QueryPerformanceCounter Lib "kernel32" (lpPerformanceCount As LARGE_INTEGER) As Long
Public Declare Function QueryPerformanceFrequency Lib "kernel32" (lpFrequency As LARGE_INTEGER) As Long
Public Type LARGE_INTEGER
    lowpart As Long
    highpart As Long
End Type

Public GLO As New Globals8
'-----------
' Shaders
'------------
Public ClipPlane(5) As Boolean



Public Type ASMShader
  Func() As Long
  Size As Long
  Handle As Long
  Name As String
End Type
Public PixelShader() As ASMShader, NbPixelShader As Integer
Public VertexShader() As ASMShader, NbVertexShader  As Integer

Public Type TVPlane
  normal As D3DVECTOR
  offset As Single
End Type

Public Enum CONST_TV_CLASSTYPE
  TV_ENT_PLAYER = 1
  TV_ENT_DEATHMATCH
  TV_ENT_MODEL
  TV_ENT_LIGHT
  TV_ENT_SPRITE
  TV_ENT_USER
  TV_ENT_OTHER
End Enum

Public Type Entity
   EntityName As String
   Target As String
   Origin As D3DVECTOR
   Direction As D3DVECTOR
   angle As D3DVECTOR2
   ClassName As String
   ClassType As CONST_TV_CLASSTYPE
   MeshAssocied As Integer
   ActorAssocied As Integer
   User() As String
   Class() As String
End Type

Public Type LensFlare
    Vertex(4) As D3DLVERTEX
    Texture As Integer
    Size As Single
    Position As Single
End Type

Public LensFlares() As LensFlare, NumLensFlare As Integer

Public Type ShaderLayer
   tex() As Integer
   NbTex As Integer
   FPS As Single
   
   Flags As Long 'Flags for this layer
   
   Key As Single
   KeyRgb As Single
   KeyTc As Single
   
   SrcBlend As Long
   DestBlend As Long
   MapType As Integer  '0: Normal map texture (1 texture)
                       '1: Lightmap *
                       '2: Animation textures
                       '3: Invisible
                       
   RgbColor As Integer
   RgbColorTransform As Integer
   RgbValue As D3DCOLORVALUE
   
   TcCoord As Integer
   TcCoordTransform As Integer
   TcCoordScale As D3DVECTOR2
   TcCoordScroll As D3DVECTOR2
   TcCoordGen As Integer
   
   Material As Integer
   
End Type

Public Const SHADER_BLEND = 1
Public Const SHADER_TCMOD = 2
Public Const SHADER_DEPTHWRITE = 4
Public Const SHADER_ANIMMAP = 8
Public Const SHADER_LIGHTMAP = 16
Public Const SHADER_ALPHATEST = 32
Public Const SHADER_TCMODSCALE = 64
Public Const SHADER_TCMODSCROLL = 128
Public Const SHADER_TCMODTURB = 256
Public Const SHADER_SKY = 512

Public Type Shader
   Layers() As ShaderLayer
   NbLayer As Integer
   ShaderName As String
   Flags As Long
   Object As Shader8
   Active As Boolean
End Type

Public Shaders() As Shader
Public NbShaders As Long

Public Type Terrain
   Position As D3DVECTOR
   Matrix As D3DMATRIX
   RefMatrix As D3DMATRIX
   Vertex() As D3DVERTEX2
   NIndices As Long
   NVertex As Long
   Precision As Single
   width As Long
   height As Long
   TextureI As Long
   MaterialI As Long
   LightmapI As Long
   Name As String
   BoxMin As D3DVECTOR
   BoxMax As D3DVECTOR
   Sphere As D3DVECTOR
   Radius As Single
   Mesh As D3DXMesh
   Prog As D3DXPMesh
   Adj As D3DXBuffer
   AttributeTable() As Long
   Enabled As Boolean
End Type

Public Type WaterEffect
 BumpMatrix As D3DMATRIX
 Plane As D3DPLANE
 Reflected As Boolean
 'BumpMapped As Boolean
 Altitude As Single
 AlphaBlended As Boolean
 AlphaFactor As Single
 DestBlend As Long
 SrcBlend As Long
 Enabled As Boolean
 Texture As Integer
 Animated As Boolean
 AnimKey As Single
 TextureScale As Single
 Vertex(4) As D3DLVERTEX2
 WaterBump As Direct3DTexture8
End Type

Public Type CloudLayer
  Material As Integer
  Texture As Integer
  CVertex(4) As D3DVERTEX2
  Moving As Boolean
  VelocityX As Single
  VelocityZ As Single
  TilingX As Single
  TilingZ As Single
  PosX As Single
  PosZ As Single
  
End Type

Public Type LandScapeDay
 SunVertex(4) As D3DVERTEX2
 SunMaterial As Integer
 Layers() As CloudLayer
 CloudFlags As Integer
 TimeDay As Single
 LandscapeCloud As Boolean
 LandscapeSun As Boolean
 sunposition As D3DVECTOR
 SunTexture As Integer
 LandscapeRain As Boolean
End Type

Public Type TextureLayer
  Texture As Integer
  power As Single
End Type

Public Type LandscapeType
  Mode As Integer
  Terrain() As Terrain
  NbTerrain As Integer
  YFactor As Single
  Water As WaterEffect
  DayFX As LandScapeDay
  LandscapeMode As Integer
  TexUsed() As Integer
  TexUsedNumber As Integer
  TexWidth As Single
  TexHeight As Single
  DetailWidth As Single
  DetailHeight As Single
  DetailTexture As Integer
  LayerTex() As TextureLayer
  NumLayerTex As Integer
  ShadowMap As Integer
End Type


Public Type LightType
   Light As D3DLIGHT8
   Active As Boolean
   Enabled As Boolean
   LightName As String
End Type

Public Light(100) As LightType
Public NBLight As Long


Public Type SoundType
   SoundBuffer As DirectSoundSecondaryBuffer8
   SoundName As String
   Playing As Boolean
   Active As Boolean
End Type

Public PrimaryBuffer As DirectSoundPrimaryBuffer8
Public DirectSound As DirectSound8
Public Sounds() As SoundType
Public NBSounds As Integer

Public Type MeshGroup
  Indices() As Long
  IndicesCount As Long
  VertexOffset As Long
  TextureI As Integer
  MaterialI As Integer
  LightmapI As Integer
  AlphaBlending As Integer
  ShpereMapping As Integer
  ShaderI As Integer
  ShaderKey As Single
  ShaderTCKey As Single
  ShaderCKey As Single
End Type

Public Type MeshSettings
  Position As D3DVECTOR
  Rotation As D3DVECTOR
  Scaling As D3DVECTOR
  color As Long
  VertexBufferEnabled As Boolean
  ProgressiveMesh As Boolean
  CurrentVertexNumber  As Long
  
  TextureI As Long
  MaterialI As Long
  Changed As Boolean
  SphereMapping As Boolean
  SphereMappingMode As Long
  Visible As Long
  
  Billboard As Boolean
  SPREnabled As Boolean
  SPRFrame As Integer
  SPRNumFrame As Integer
  SPROffset As Integer
  
  SphereClip As D3DVECTOR
  SphereRadius As Single
  SphereComputed As Boolean
  
  BoxMin As D3DVECTOR
  BoxMax As D3DVECTOR
  
  Velocity As D3DVECTOR
  Gravity As Single
End Type
  

Public Type MeshFeatures
  Settings As MeshSettings
  Vertex() As D3DVERTEX2
  LVertex() As D3DLVERTEX2
  NbVertex As Long
  TypeVertex As Integer
  NumGroup As Integer
  Groups() As MeshGroup
  PrimType As Integer
End Type

Public NumFaceDisplayed As Long

Public Type MeshType
   MeshName As String
   Active As Boolean
   Enabled As Boolean
   Changed As Boolean
   Created As Boolean
   Object As Mesh8
   Adj As D3DXBuffer
   D3DMesh As D3DXMesh
   ProgMesh As D3DXPMesh
   D3DMeshSkin As D3DXSkinMesh
   Features As MeshFeatures
   
   Matrix As D3DMATRIX
   Translation As D3DMATRIX
   Rotation As D3DMATRIX
   
   NumChild As Integer
   Childs() As Integer
   Parent As Integer
   
   CullMode As Integer
   Modified As Boolean
   DeathTime As Single
End Type


'---------------------------
' TVO Landscape
'---------------------------
Public Type TvoEDLand
  NbTerrain As Integer
  
End Type


Public Type TvoED
  Ident As String * 4
  version As Integer
  DataOffset As Long
End Type



Public SphereMappingMatrix As D3DMATRIX
Public CollisionNormal As D3DVECTOR, CollisionImpact As D3DVECTOR


Public Mesh() As MeshType, NBMesh As Integer

Public NumSpriteTex As Integer, SpriteTexName() As String


Public Type TVOTextureTable
 Filename As String
 TextureName As String
 Colorkey As Long
 width As Long
 height As Long
End Type

Public ChunkString(1000) As String

Global Const CHUNK_VERSION_TV = 8

Public Enum ChunkEnum
   CHUNK_VERSION = 1  'long (must be 8)
   
   'Mesh chunks
   CHUNK_MESH = 2     'indicates that the following stuff is a mesh
   CHUNK_FEATURES = 3 'this CHUNK is composed of
                      '  D3DVECTOR  Position
                      '  D3DVECTOR  Rotation
                      '  D3DVECTOR  Scaling
                      '  D3DVECTOR  SphereCenter (for clipping)
                      '  Single     SphereRadius
                      '
   CHUNK_NBFACE = 5   'Long representing the face number (or group number)
   CHUNK_VERTEX = 6   'A array of D3DVERTEX2 values (if the mesh is composed of NORMALVERTEX)
   CHUNK_NBVERTEX = 7 'Long representing the vertexes number of the mesh
   CHUNK_SPRITE = 8   'indicates that the mesh is a SPR sprite (no content)
   CHUNK_LVERTEX = 11 'A array of D3DLVERTEX2 value (if the mesh is composed of COLORVERTEX)
   CHUNK_PRIMITIVETYPE = 12 'A Long that represents the Primitive used (CONST_TV_PRIMITIVETYPE value)
   CHUNK_BILLBOARD = 13 'indicates that the mesh is a billboard. (no content)
   CHUNK_MESHNAME = 25 'the name of the mesh follows this chunk header
   CHUNK_ENDMESH = 16 '
   
   
   
   'Start face desc
   CHUNK_FACE = 4     'indicates that the following stuff is a face description
   'Face options chunks
   CHUNK_SPHEREMAPPING = 14  '+ Int SphereMapping Mode
   CHUNK_ALPHABLENDING = 25 ' + Int AlphaBlending mode (CONST_TV_BLENDINGMODE)
   CHUNK_FACETEXTURE = 17 '+ Int Texture Index in Textures table of the group/face
   CHUNK_FACEMATERIAL = 18 ' + Int Material Index in Textures table of the Group/facde
   CHUNK_FACELIGHTMAP = 19 ' + Int LightMap Index (not used usally)
   CHUNK_FACEVERTEXSTART = 20 '+ Long Vertex offset in the vertexes array
   CHUNK_FACEVERTEXNUMBER = 21 '+ Long Vertex number of the face/group
   
   'End face.
   CHUNK_ENDFACE = 15
   
   'Texture table/factory
   CHUNK_TEXTURE = 9  'Textures table with filenames & colorkey
                      'Array of Textures
   CHUNK_NBTEXTURE = 10 'long representing the numbers of textures used by this mesh
   
   'Material table/factory
   CHUNK_MATERIAL = 22  ' Materail Table with mat name & D3DMATERIAL8
   CHUNK_NBMATERIAL = 23 'Materials number
   
   
   'Particles System
   CHUNK_PARTICLESYSTEM = 200
   CHUNK_STANDARDFORMAT = 201  'This chunk must be followed of the BS Structure type
   
   'Track System
   CHUNK_TRACKSYSTEM = 400
   CHUNK_TRACKNUMNODES = 401
   
   'Landscape System
   CHUNK_LANDSCAPE = 500
   CHUNK_LANDFEATURES1 = 501  'Land features :
                              '  int NB Terrain
   CHUNK_TERRAIN = 510        'Start a terrain desc
   CHUNK_TERRAINFEATURES = 511 ' Terrain Desc :
                               '   D3DVECTOR : Terrain Position
                               '   long : Terrain width
                               '   long : terrain height
                               '   long : terrain texture
                               '   long : terrain material
   CHUNK_TERRAINFEATURES2 = 512 ' Terrain Desc 2 (Sphere)
                                ' D3DVECTOR : Sphere
                                ' single Sphere radius
   CHUNK_TERRAINHEIGHTMAP = 515 ' This chunk is followed by (width*height) singles value representing the altitude
   
   
   CHUNK_LANDSCAPE_WATER = 512  'Chunk indicates that the water is enabled. This chunk is followed by :
   CHUNK_STANDARDWATERFORMAT = 513  'the standard water structure type
   
   CHUNK_GENERALNAME = 1000   ' A global name chunk for all the classes.
   
End Enum

Public Type ChunkMesh
  ChunkType As Integer
  ChunkLength As Long
  ChunkOffset As Long
End Type

Public Enum CONST_TV_PLANES
   TV_PLANE_NEAR = 1
   TV_PLANE_FAR = 2
   TV_PLANE_LEFT = 4
   TV_PLANE_RIGHT = 8
   TV_PLANE_TOP = 16
   TV_PLANE_BOTTOM = 32
End Enum
   
Sub AddChunk(ChunkType As Integer, C() As ChunkMesh, Optional tff As Integer = 2)
 ReDim Preserve C(UBound(C) + 1)
 Static last As Long
 Put #tff, , ChunkType
 C(UBound(C)).ChunkOffset = Seek(tff)
 C(UBound(C)).ChunkType = ChunkType
 C(UBound(C) - 1).ChunkLength = Seek(tff) - last - 2
 last = Seek(tff)

End Sub




Function CreateShader(SName As String) As Integer
   Dim i As Integer
   For i = 1 To NbShaders
     If SName = Shaders(i).ShaderName Or Shaders(i).Active = False Then
       Exit For
     End If
   Next i
   
   If i = NbShaders + 1 Then
      NbShaders = NbShaders + 1
      ReDim Preserve Shaders(NbShaders)
   End If
   
   Shaders(i).ShaderName = SName
   Shaders(i).NbLayer = 0
   Shaders(i).Flags = 0
   ReDim Shaders(i).Layers(0)
   Set Shaders(i).Object = New Shader8
   Shaders(i).Object.SetShaderIndex i
  ' Shaders(i).Layers(1).FPS = 15
   Shaders(i).Active = True
   
   
   CreateShader = i
   
End Function

Public Function GetArg(Str As String, ArgArray() As String) As Integer
          Dim ani$, inst As Long, nspl As Long
          ani$ = Str
          nspl = 0
          Do
            inst = InStr(ani$, " ")
            If ani$ = "" Then Exit Do
            If inst = 0 Then
             'ReDim Preserve Spl(nspl)
             ArgArray(nspl) = Trime(ani$)
             nspl = nspl + 1
             ani$ = ""
            Else
             'ReDim Preserve Spl(nspl)
             ArgArray(nspl) = Left$(ani$, inst - 1)
             nspl = nspl + 1
             ani$ = mid$(ani$, inst + 1)
            End If
          Loop Until ani$ = ""
          GetArg = nspl
 
End Function

Sub InitChunkString()
  ChunkString(1) = "CHUNK_VERSION"
  ChunkString(2) = "CHUNK_MESH"
  ChunkString(3) = "CHUNK_FEATURES"
  ChunkString(4) = "CHUNK_FACE"
  ChunkString(5) = "CHUNK_NBFACE"
  ChunkString(6) = "CHUNK_VERTEX"
  ChunkString(7) = "CHUNK_NBVERTEX"
  ChunkString(8) = "CHUNK_SPRITE"
  ChunkString(9) = "CHUNK_TEXTURE"
  ChunkString(10) = "CHUNK_NBTEXTURE"
  ChunkString(11) = "CHUNK_LVERTEX"
  ChunkString(12) = "CHUNK_PRIMITIVETYPE"
  ChunkString(13) = "CHUNK_BILLBOARD"
  ChunkString(14) = "CHUNK_SPHEREMAPPING"
  ChunkString(15) = "CHUNK_ENDFACE"
  ChunkString(16) = "CHUNK_ENDMESH"
  ChunkString(17) = "CHUNK_FACETEXTURE"
  ChunkString(18) = "CHUNK_FACEMATERIAL"
  ChunkString(19) = "CHUNK_FACELIGHTMAP"
  ChunkString(20) = "CHUNK_FACEVERTEXSTART"
  ChunkString(21) = "CHUNK_FACEVERTEXNUMBER"
  ChunkString(22) = "CHUNK_MATERIAL"
  ChunkString(23) = "CHUNK_NBMATERIAL"
  ChunkString(24) = ""
  ChunkString(25) = "CHUNK_MESHNAME"
  ChunkString(200) = "CHUNK_PARTICLESYSTEM"
  ChunkString(201) = "CHUNK_STANDARDFORMAT"
  ChunkString(400) = "CHUNK_TRACKSYSTEM"
  ChunkString(401) = "CHUNK_TRACKNUMNODES"
  ChunkString(500) = "CHUNK_LANDSCAPE"
  ChunkString(501) = "CHUNK_LANDFEATURES1"
  ChunkString(510) = "CHUNK_TERRAIN"
  ChunkString(411) = "CHUNK_TERRAINFEATURES"
  ChunkString(515) = "CHUNK_TERRAINHEIGHTMAP"
  ChunkString(512) = "CHUNK_LANDSCAPE_WATER"
  ChunkString(513) = "CHUNK_STANDARDWATERFORMAT"
End Sub

Sub LoadMeshFromTVO(Filename As String, MI As Integer, TextureToo As Boolean)
 InitChunkString
 Open Filename For Binary As 2
   'Parse TVO chunk
   Dim magic As String * 4
   Dim C() As ChunkMesh
   Dim nbchunk As Long, i As Long, z As Long
   Get #2, , magic
   If magic <> "TVOF" Then
     AddLog "This file isn't a good TrueVision Format file."
     Exit Sub
   End If
   Dim TableOffset As Long, TableLength As Long
   Dim version As Long
        
   Get #2, , TableOffset
   
   Seek #2, TableOffset
   nbchunk = (LOF(2) - TableOffset) / 10
   ReDim Preserve C(nbchunk)
   Get #2, , C()
   For i = 1 To nbchunk
    Debug.Print ChunkString(C(i).ChunkType), C(i).ChunkLength, C(i).ChunkOffset
   Next i
        
   Dim Rotation As D3DVECTOR, Position As D3DVECTOR, Scaling As D3DVECTOR
   Dim MeshStart As Boolean, NbVertex As Long
        
   Dim Face As Integer, NTex As Integer
   Face = 0
   With Mesh(MI)
   For i = 1 To nbchunk
   
     Seek #2, C(i).ChunkOffset
    
     Select Case C(i).ChunkType
       Case CHUNK_VERSION
        Get #2, , version
        Debug.Print "Version "; version
       Case CHUNK_FEATURES
        Get #2, , Mesh(MI).Features.Settings.Position
        Get #2, , Mesh(MI).Features.Settings.Rotation
        Get #2, , Mesh(MI).Features.Settings.Scaling
       Case CHUNK_MESH
         MeshStart = True
       Case CHUNK_MESHNAME
        Get #2, , Mesh(MI).MeshName
       Case CHUNK_NBVERTEX
         Get #2, , NbVertex
         Debug.Print "Nb Vertex :"; NbVertex
       Case CHUNK_VERTEX
         If NbVertex <> 0 And MeshStart = True Then
           ReDim Preserve Mesh(MI).Features.Vertex(NbVertex)
           Get #2, , Mesh(MI).Features.Vertex
         End If
         
       Case CHUNK_LVERTEX
         If NbVertex <> 0 And MeshStart = True Then
           ReDim Preserve Mesh(MI).Features.LVertex(NbVertex)
           Get #2, , Mesh(MI).Features.LVertex
           .Features.TypeVertex = TV_COLOR_VERTEX
         End If
         
       Case CHUNK_PRIMITIVETYPE
         Get #2, , Mesh(MI).Features.PrimType
       Case CHUNK_BILLBOARD
          .Features.Settings.Billboard = True
       Case CHUNK_SPRITE
          .Features.Settings.SPREnabled = True
       Case CHUNK_NBFACE
          Get #2, , .Features.NumGroup
          ReDim .Features.Groups(.Features.NumGroup)
          .Object.SetGroupPointer .Features.NumGroup
       
       Case CHUNK_FACE
          Face = Face + 1
          
          'Face stuff
          Case CHUNK_FACEVERTEXSTART
           Get #2, , .Features.Groups(Face).VertexOffset
          Case CHUNK_FACELIGHTMAP
           Get #2, , .Features.Groups(Face).LightmapI
          Case CHUNK_FACEVERTEXNUMBER
           Get #2, , .Features.Groups(Face).IndicesCount
          Case CHUNK_FACETEXTURE
           Get #2, , .Features.Groups(Face).TextureI
          Case CHUNK_FACEMATERIAL
           Get #2, , .Features.Groups(Face).MaterialI
          Case CHUNK_ALPHABLENDING
           Get #2, , .Features.Groups(Face).AlphaBlending
          Case CHUNK_SPHEREMAPPING
           Get #2, , .Features.Groups(Face).ShpereMapping
       Case CHUNK_ENDFACE
       
       Case CHUNK_ENDMESH
         MeshStart = False
         
       Case CHUNK_NBTEXTURE
           Get #2, , NTex
       Case CHUNK_TEXTURE
           If NTex > 0 Then
             Dim TF As New TextureFactory8
             For z = 1 To NTex
               
             Next z
           End If
          
     End Select
   Next i
 End With
 
 Close #2
End Sub

Function LoadSimpleShader(ttf As Integer, Name As String, Optional Quake3 As Boolean) As Integer
  Dim si As Long
  si = CreateShader(Name)
'  On Error Resume Next
  Dim M As New MaterialFactory8
  On Error Resume Next
  LoadSimpleShader = si
  With Shaders(si)
  
  Dim Ident As Long
  Dim lin$, ins As Long, inst As Long
  Dim keyw$, ani$, rest$, lef$, num As Long
  Dim i As Long, lp As Long
  Dim LE$, ri$
  
  .NbLayer = 0
  
  Ident = 0
  Do
   Line Input #ttf, lin$
   lin$ = Trime(lin$)
   
   'Ignore // and empty line
   If Left$(lin$, 2) = "//" Or lin$ = "" Then
     GoTo ChangeLine
   End If
    
   If lin$ = "{" Then
     Ident = Ident + 1
     If Ident = 2 Then
       .NbLayer = .NbLayer + 1
       ReDim Preserve .Layers(.NbLayer)
       .Layers(.NbLayer).MapType = 0
       .Layers(.NbLayer).DestBlend = 0
       .Layers(.NbLayer).SrcBlend = -1
       .Layers(.NbLayer).TcCoord = 0
       .Layers(.NbLayer).Flags = 0
       .Layers(.NbLayer).RgbColor = 0
       .Layers(.NbLayer).NbTex = 1
       .Layers(.NbLayer).Material = M.CreateMaterialQuick(1, 1, 1, 1, "MatShader")
       'Device.SetRenderState D3DRS_AMBIENTMATERIALSOURCE
       ReDim Preserve .Layers(.NbLayer).tex(1)
       .Layers(.NbLayer).tex(1) = -1  'Default texture
     End If
   End If
   
     ins = InStr(lin$, " ")
       If ins > 0 Then
         keyw$ = Trime(Left$(lin$, ins - 1))
         rest$ = Trime(mid$(lin$, InStr(lin$, " ") + 1))
       Else
         keyw$ = lin$
         rest$ = ""
       End If
   If Ident = 1 Then
    Select Case UCase$(keyw$)
      Case "SURFACEPARM"
        If Trim$(UCase$(rest$)) = "SKY" Then
        .Flags = SHADER_SKY
        End If
    End Select
   End If
   If Ident = 2 Then
      Select Case UCase$(keyw$)
        Case "MAP"
         .Layers(.NbLayer).MapType = 0
         .Layers(.NbLayer).NbTex = 1
         ReDim .Layers(.NbLayer).tex(1)
         If Trim$(UCase$(rest$)) = "$LIGHTMAP" Then
           .Layers(.NbLayer).tex(1) = -2
           .Layers(.NbLayer).Flags = .Layers(.NbLayer).Flags Or SHADER_LIGHTMAP
           .Layers(.NbLayer).MapType = 1 'lightmap
         ElseIf Trim$(UCase$(rest$)) = "$TEXTURE" Then
           .Layers(.NbLayer).tex(1) = -1     'default texture (set by SetTexture func)
         Else
          If Quake3 = False Then
            .Layers(.NbLayer).tex(1) = GetTex(Trim$(rest$)).GetTexIndex
          Else
           If Exist(rest$) = False Then
               Mid$(rest$, Len(rest$) - 3, 4) = ".JPG"
           End If
          .Layers(.NbLayer).tex(1) = LocTexFactory.LoadTexture(rest$, rest$, , , TV_COLORKEY_BLACK).GetTexIndex
          End If
         End If

        Case "DEPTHWRITE"
          .Layers(.NbLayer).Flags = .Layers(.NbLayer).Flags Or SHADER_DEPTHWRITE
          
        Case "ANIM"
          .Layers(.NbLayer).Flags = .Layers(.NbLayer).Flags Or SHADER_ANIMMAP
          .Layers(.NbLayer).MapType = 2
          ani$ = UCase$(Trim$(rest$))
          inst = InStr(ani$, " ")
          .Layers(.NbLayer).FPS = val(Left$(ani$, inst - 1))
          ani$ = mid$(ani$, inst + 1)
           .Layers(.NbLayer).NbTex = 0
          Do While inst <> 0
           inst = InStr(ani$, " ")
           If inst = 0 Then Exit Do
                    
           .Layers(.NbLayer).NbTex = .Layers(.NbLayer).NbTex + 1
           ReDim Preserve .Layers(.NbLayer).tex(.Layers(.NbLayer).NbTex)
           If Quake3 = False Then
           .Layers(.NbLayer).tex(.Layers(.NbLayer).NbTex) = GetTex(Trim$(Left$(ani$, inst - 1))).GetTexIndex
           Else
            lef$ = Trim$(Left$(ani$, inst - 1))
           .Layers(.NbLayer).tex(.Layers(.NbLayer).NbTex) = LocTexFactory.LoadTexture(lef$, lef$, , , TV_COLORKEY_BLACK).GetTexIndex
           End If
           ani$ = mid$(ani$, inst + 1)
          Loop
         ' .Layers(.NbLayer).NbTex = .Layers(.NbLayer).NbTex - 1
        Case "TCMOD"
          ' Texture coordintates modify
          ' Used by texture coords transform matrix.
          ani$ = UCase$(Trime(rest$))
          Dim argu(10) As String
          num = GetArg(ani$, argu())
          
          'MOD type
          .Layers(.NbLayer).Flags = .Layers(.NbLayer).Flags Or SHADER_TCMOD
          Select Case argu(0)
            Case "SCALE"
            .Layers(.NbLayer).Flags = .Layers(.NbLayer).Flags Or SHADER_TCMODSCALE
              .Layers(.NbLayer).TcCoordScale = Vec2(val(argu(1)), val(argu(2)))
            Case "SCROLL"
            .Layers(.NbLayer).Flags = .Layers(.NbLayer).Flags Or SHADER_TCMODSCROLL
              .Layers(.NbLayer).TcCoordScroll = Vec2(val(argu(1)), val(argu(2)))
          End Select
        Case "ANIMMAP"
          
          .Layers(.NbLayer).Flags = .Layers(.NbLayer).Flags Or SHADER_ANIMMAP
          .Layers(.NbLayer).MapType = 2
          ani$ = UCase$(Trime(rest$))
          Dim Spl() As String, nspl As Long
          nspl = 0
          
          'Spl = Split(ani$)
          Do
            inst = InStr(ani$, " ")
            If ani$ = "" Then Exit Do
            If inst = 0 Then
             ReDim Preserve Spl(nspl)
             Spl(nspl) = Trime(ani$)
             nspl = nspl + 1
             ani$ = ""
            Else
             ReDim Preserve Spl(nspl)
             Spl(nspl) = Left$(ani$, inst - 1)
             nspl = nspl + 1
             ani$ = mid$(ani$, inst + 1)
            End If
          Loop Until ani$ = ""
          num = UBound(Spl)
          .Layers(.NbLayer).NbTex = num - 1
          ReDim Preserve .Layers(.NbLayer).tex(num - 1)
          .Layers(.NbLayer).FPS = val(Spl(0))
          For i = 1 To num - 1
            If Exist(Spl(i)) = False Then Mid(Spl(i), Len(Spl(i)) - 3) = ".JPG"
             .Layers(.NbLayer).tex(i) = LocTexFactory.LoadTexture(CStr(Spl(i)), CStr(Spl(i)), , , TV_COLORKEY_BLACK).GetTexIndex
          Next i
        Case "BLEND"
         .Layers(.NbLayer).Flags = .Layers(.NbLayer).Flags Or SHADER_BLEND
         Select Case UCase$(Trime(rest$))
           Case "ADD"
             .Layers(.NbLayer).DestBlend = D3DBLEND_ONE
             .Layers(.NbLayer).SrcBlend = D3DBLEND_ONE
           Case "ALPHA"
             .Layers(.NbLayer).DestBlend = D3DBLEND_INVSRCALPHA
             .Layers(.NbLayer).SrcBlend = D3DBLEND_SRCALPHA
           Case "LIGHTMAP"
             .Layers(.NbLayer).DestBlend = D3DBLEND_ZERO
             .Layers(.NbLayer).SrcBlend = D3DBLEND_DESTCOLOR
           Case "COLOR"
             .Layers(.NbLayer).SrcBlend = D3DBLEND_SRCCOLOR
             .Layers(.NbLayer).DestBlend = D3DBLEND_ONE
         End Select
       
       
       Case "BLENDFUNC"
           .Layers(.NbLayer).Flags = .Layers(.NbLayer).Flags Or SHADER_BLEND
           rest$ = Trime(rest$)
           lp = InStr(rest$, " ")
           If lp = 0 Then
            LE$ = rest$
           Else
            LE$ = Left$(rest$, lp - 1)
            ri$ = mid$(rest$, lp + 1)
           End If
          With .Layers(.NbLayer)
            Select Case UCase$(LE$)
              Case "BLEND"
               .SrcBlend = D3DBLEND_ONE
               .DestBlend = D3DBLEND_ONE
              Case "FILTER"
               .SrcBlend = D3DBLEND_DESTCOLOR
               .DestBlend = D3DBLEND_ZERO
              Case "ONE", "GL_ONE"
               .SrcBlend = D3DBLEND_ONE
              Case "ZERO", "GL_ZERO"
               .SrcBlend = D3DBLEND_ZERO
              Case "COLOR", "GL_SRC_COLOR"
               .SrcBlend = D3DBLEND_SRCCOLOR
              Case "DESTCOLOR", "GL_DST_COLOR"
               .SrcBlend = D3DBLEND_DESTCOLOR
              Case "ALPHA", "GL_SRC_ALPHA"
               .SrcBlend = D3DBLEND_SRCALPHA
              Case "INVALPHA", "GL_ONE_MINUS_SRC_ALPHA"
               .SrcBlend = D3DBLEND_INVSRCALPHA
             End Select
             Select Case UCase$(ri$)
              Case "ONE", "GL_ONE"
               .DestBlend = D3DBLEND_ONE
              Case "ZERO", "GL_ZERO"
               .DestBlend = D3DBLEND_ZERO
              Case "COLOR", "GL_DST_COLOR "
               .DestBlend = D3DBLEND_DESTCOLOR
              Case "SRCCOLOR", "GL_DST_COLOR"
               .DestBlend = D3DBLEND_SRCCOLOR
              Case "ALPHA", "GL_DST_ALPHA"
               .DestBlend = D3DBLEND_SRCALPHA
              Case "INVALPHA", "GL_ONE_MINUS_SRC_ALPHA"
               .DestBlend = D3DBLEND_INVSRCALPHA
             End Select
          End With
      End Select
   
       
   
   End If
   
   If lin$ = "}" Then
     Ident = Ident - 1
   End If
ChangeLine:
  Loop Until lin$ = "}" And Ident = 0
  End With
End Function
Public Function ShaderArray(o As Integer, ParamArray Arr() As Variant) As Long

End Function

Sub PixelShaderInit()
    Dim Pointer As Long, i As Long
    Dim Buffer As D3DXBuffer
    Dim bd()
    Set Buffer = D3DX.AssembleShaderFromFile(App.Path + "\lightmap.asm", D3DXASM_SKIPVALIDATION, ErrStr, Nothing)
    Debug.Print Buffer.GetBufferSize / 4, ErrStr
    Pointer = Buffer.GetBufferPointer
    
    Dim b() As Long
    ReDim b(Buffer.GetBufferSize / 4)
    Dim by() As Byte
    ReDim by(Buffer.GetBufferSize)
    DXCopyMemory b(0), ByVal Buffer.GetBufferPointer, Buffer.GetBufferSize
    DXCopyMemory by(0), ByVal Buffer.GetBufferPointer, Buffer.GetBufferSize
   
   For i = 0 To Buffer.GetBufferSize / 4
    Debug.Print Hex$(b(i))
   Next i
   For i = 1 To Buffer.GetBufferSize
    Debug.Print Chr(by(i));
   Next i
 Dim PixelShader As Long
 PixelShader = Device.CreatePixelShader(b(0))
 
 Dim t() As Long
 bd = Array(&HFFFF0100, _
           &H9FFFE, &H58443344, &H68532038, &H72656461, &H73734120, &H6C626D65, &H56207265, &H69737265, &H30206E6F, &H31392E, _
           &H42, &HB00F0000, _
           &H42, &HB00F0001, _
           &H1, &H800F0000, &HB0E40000, _
           &H5, &H800F0000, &H80E40000, &HB0E40001, _
           &H2, &H800F0000, &H80E40000, &H90E40000, _
           &HFFFF)
           
           

End Sub
Sub RenderGroup(MI As Integer, Group As Integer, Texture As Integer, Material As Integer)
        Dim V1 As D3DLVERTEX2
        Dim V As D3DVERTEX2
        With Mesh(MI)
           If .Features.Groups(Group).IndicesCount < 2 Then Exit Sub
           If Texture > 0 Then
             Device.SetTexture 0, Textures(Texture).TexSurf
          Else
             Device.SetTexture 0, Nothing
          End If
          If Material > 0 Then
             Device.SetMaterial Materials(Material).Mat
          Else
             Device.SetMaterial Materials(0).Mat
          End If
          
          If .Features.Settings.VertexBufferEnabled = True Then
            If .Features.Settings.ProgressiveMesh = True Then
            .ProgMesh.DrawSubset Group - 1
            Else
            .D3DMesh.DrawSubset Group - 1
            End If
          Else

           If .Features.PrimType = D3DPT_TRIANGLELIST Then
           
            NumFaceDisplayed = NumFaceDisplayed + .Features.Groups(Group).IndicesCount \ 3
            Device.SetVertexShader .Features.TypeVertex
            If .Features.TypeVertex <> TV_COLOR_VERTEX Then
              Device.DrawPrimitiveUP D3DPT_TRIANGLELIST, .Features.Groups(Group).IndicesCount \ 3, .Features.Vertex(.Features.Groups(Group).VertexOffset), Len(V)
            Else
              Device.DrawPrimitiveUP D3DPT_TRIANGLELIST, .Features.Groups(Group).IndicesCount \ 3, .Features.LVertex(.Features.Groups(Group).VertexOffset), Len(V1)
            End If
           Else
           NumFaceDisplayed = NumFaceDisplayed + .Features.Groups(Group).IndicesCount - 2
           Device.SetVertexShader .Features.TypeVertex
            If .Features.TypeVertex <> TV_COLOR_VERTEX Then
              Device.DrawPrimitiveUP D3DPT_TRIANGLESTRIP, .Features.Groups(Group).IndicesCount - 2, .Features.Vertex(.Features.Groups(Group).VertexOffset), Len(V)
            Else
              Device.DrawPrimitiveUP D3DPT_TRIANGLESTRIP, .Features.Groups(Group).IndicesCount - 2, .Features.LVertex(.Features.Groups(Group).VertexOffset), Len(V1)
           End If
          End If
         End If

        End With
End Sub

Sub RenderMesh(MI As Integer, RenderPart As Boolean, Optional StartGroup As Long = 1, Optional EndGroup As Long = 0)
    Dim i As Long
    Dim ScaleMatrix As D3DMATRIX, TranslationMatrix As D3DMATRIX
    'On Error Resume Next
    Dim SGroup As Long, EGroup As Long
    
    With Mesh(MI)
     If .Enabled = False Then Exit Sub
     If .Features.NbVertex <= 0 Then Exit Sub
     If .DeathTime > 0 Then
       .DeathTime = .DeathTime - SystemFrame
       If .DeathTime <= 0 Then
        .Enabled = False
        Exit Sub
       End If
     End If
       Device.SetTransform D3DTS_WORLD, .Matrix
      If .Features.Settings.SphereComputed = True And .Changed = False Then
         If CheckSphere(.Features.Settings.SphereClip, .Features.Settings.SphereRadius) = 0 Then
           Exit Sub
         End If
      End If
      If RenderPart = False Then
        SGroup = 1
        EGroup = .Features.NumGroup
      Else
        SGroup = StartGroup
        EGroup = EndGroup
        If EndGroup = 0 Then EGroup = .Features.NumGroup
      End If
     If .Changed = True Or Mesh(MI).Features.Settings.Billboard = True Or .Parent <> 0 Then
       D3DXMatrixIdentity .Matrix
       D3DXMatrixScaling ScaleMatrix, .Features.Settings.Scaling.x, .Features.Settings.Scaling.y, .Features.Settings.Scaling.z
       D3DXMatrixTranslation TranslationMatrix, .Features.Settings.Position.x, .Features.Settings.Position.y, .Features.Settings.Position.z
       If Mesh(MI).Features.Settings.Billboard = False Then
       ScaleMatrix = MRotate(ScaleMatrix, .Features.Settings.Rotation.x, .Features.Settings.Rotation.y, .Features.Settings.Rotation.z)
       Else
       'ScaleMatrix = MRotate(ScaleMatrix, 0, GetCameraAngle, 0)
       ScaleMatrix = MRotate(ScaleMatrix, 0, GetCameraAngle, 0)
       
       End If
       D3DXMatrixMultiply .Matrix, ScaleMatrix, TranslationMatrix
       If .Parent <> 0 Then
        D3DXMatrixMultiply .Matrix, Mesh(.Parent).Matrix, .Matrix
       End If
       Device.SetTransform D3DTS_WORLD, .Matrix
      .Changed = False
        
     End If
    
    End With
    'For i = 1 To NBMaterial
    '  With Materials(i).Mat
    '    Debug.Print .Ambient.r, .Ambient.g, .Ambient.B, .Ambient.a
    '    Debug.Print .Diffuse.r, .Diffuse.g, .Diffuse.B, .Diffuse.a
    '    Debug.Print .emissive.r, .emissive.g, .emissive.B, .emissive.a
    '    Debug.Print .Specular.r, .Specular.g, .Specular.B, .Specular.a
    '
    '  End With
    'Next i
    If Mesh(MI).Features.Settings.SphereMapping = True Then
       Device.SetTransform D3DTS_TEXTURE0, SphereMappingMatrix
       Device.SetRenderState D3DRS_CULLMODE, D3DCULL_NONE
       Device.SetTextureStageState 0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT2
       Device.SetTextureStageState 0, D3DTSS_TEXCOORDINDEX, Mesh(MI).Features.Settings.SphereMappingMode
       Device.SetRenderState D3DRS_NORMALIZENORMALS, 1
    Else
       Device.SetTransform D3DTS_TEXTURE0, WorldMatrix
       Device.SetTextureStageState 0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE
       Device.SetTextureStageState 0, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_PASSTHRU
    
    End If
    Device.SetRenderState D3DRS_CULLMODE, Mesh(MI).CullMode
   If Mesh(MI).Features.Settings.VertexBufferEnabled = True Then
  
    Device.SetRenderState D3DRS_LIGHTING, 1
    With Mesh(MI)
       For i = SGroup To EGroup
          If .Features.Groups(i).TextureI > 0 Then
             Device.SetTexture 0, Textures(.Features.Groups(i).TextureI).TexSurf
             If Textures(.Features.Groups(i).TextureI).Colorkey <> 0 Then
                    Device.SetRenderState D3DRS_ALPHATESTENABLE, 1
             End If
          Else
             Device.SetTexture 0, Nothing
          End If
          If .Features.Groups(i).MaterialI > 0 Then
             Device.SetMaterial Materials(.Features.Groups(i).MaterialI).Mat
          Else
             Device.SetMaterial Materials(0).Mat
          End If
          .D3DMesh.DrawSubset i - 1
       Next i
    End With
   Else
     'Normal vertex list
           Dim V As D3DVERTEX2, V1  As D3DLVERTEX2
     
    
     Device.SetRenderState D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA
     Device.SetRenderState D3DRS_SRCBLEND, D3DBLEND_SRCALPHA
     Device.SetRenderState D3DRS_ALPHATESTENABLE, 1
     Device.SetRenderState D3DRS_ALPHABLENDENABLE, 1
          
     With Mesh(MI)
       If .Features.TypeVertex = TV_COLOR_VERTEX Then
             Device.SetRenderState D3DRS_LIGHTING, 0
       Else
             Device.SetRenderState D3DRS_LIGHTING, 1
       End If
       
       Device.SetTransform D3DTS_WORLD, .Matrix
       For i = SGroup To EGroup
          Device.SetVertexShader .Features.TypeVertex
          
          If .Features.Groups(i).LightmapI > 0 Then
              Device.SetRenderState D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA
              Device.SetRenderState D3DRS_SRCBLEND, D3DBLEND_SRCALPHA
     
             Device.SetTexture 0, MeshLM(.Features.Groups(i).LightmapI)
              
             If .Features.PrimType = D3DPT_TRIANGLELIST Then
                 Device.SetVertexShader .Features.TypeVertex
                 If .Features.TypeVertex <> TV_COLOR_VERTEX Then
                   Device.DrawPrimitiveUP .Features.PrimType, .Features.Groups(i).IndicesCount / 3, .Features.Vertex(.Features.Groups(i).VertexOffset), Len(V)
                 Else
                   Device.DrawPrimitiveUP .Features.PrimType, .Features.Groups(i).IndicesCount / 3, .Features.LVertex(.Features.Groups(i).VertexOffset), Len(V1)
                 End If
             Else
               Device.SetVertexShader .Features.TypeVertex
               If .Features.TypeVertex <> TV_COLOR_VERTEX Then
                 Device.DrawPrimitiveUP D3DPT_TRIANGLESTRIP, .Features.Groups(i).IndicesCount - 2, .Features.Vertex(.Features.Groups(i).VertexOffset), Len(V)
               Else
                 Device.DrawPrimitiveUP D3DPT_TRIANGLESTRIP, .Features.Groups(i).IndicesCount - 2, .Features.LVertex(.Features.Groups(i).VertexOffset), Len(V1)
               End If
          End If
              Device.SetRenderState D3DRS_DESTBLEND, D3DBLEND_SRCCOLOR
              Device.SetRenderState D3DRS_SRCBLEND, D3DBLEND_ZERO
          End If
          If .Features.Groups(i).TextureI > 0 Then
             Device.SetTexture 0, Textures(.Features.Groups(i).TextureI).TexSurf
          Else
             Device.SetTexture 0, Nothing
          End If
          If .Features.Groups(i).MaterialI > 0 Then
             Device.SetMaterial Materials(.Features.Groups(i).MaterialI).Mat
          Else
             Device.SetMaterial Materials(0).Mat
          End If
          
          If .Features.PrimType = D3DPT_TRIANGLELIST Then
           Device.SetVertexShader .Features.TypeVertex
           If .Features.TypeVertex <> TV_COLOR_VERTEX Then
             Device.DrawPrimitiveUP .Features.PrimType, .Features.Groups(i).IndicesCount / 3, .Features.Vertex(.Features.Groups(i).VertexOffset), Len(V)
           Else
             Device.DrawPrimitiveUP .Features.PrimType, .Features.Groups(i).IndicesCount / 3, .Features.LVertex(.Features.Groups(i).VertexOffset), Len(V1)
           End If
          Else
          Device.SetVertexShader .Features.TypeVertex
           If .Features.TypeVertex <> TV_COLOR_VERTEX Then
             Device.DrawPrimitiveUP D3DPT_TRIANGLESTRIP, .Features.Groups(i).IndicesCount - 2, .Features.Vertex(.Features.Groups(i).VertexOffset), Len(V)
           Else
             Device.DrawPrimitiveUP D3DPT_TRIANGLESTRIP, .Features.Groups(i).IndicesCount - 2, .Features.LVertex(.Features.Groups(i).VertexOffset), Len(V1)
           End If
          End If
       Next i
    End With
   End If
   With Mesh(MI)
    If .NumChild > 0 Then
      For i = 1 To .NumChild
        RenderMesh .Childs(i), False
      Next i
    End If
   End With
End Sub

Sub RenderMesh2(MI As Integer, RenderPart As Boolean, Optional StartGroup As Long = 1, Optional EndGroup As Long = 0)
    Dim i As Integer
    Dim ScaleMatrix As D3DMATRIX, TranslationMatrix As D3DMATRIX
    Dim SGroup As Long, EGroup As Long
    Dim tex As Integer, Mat As Integer
    Dim a As Long
    
    With Mesh(MI)
     ' If this mesh isn't enabled, don't render it
     If .Enabled = False Then Exit Sub
     
     ' If there is no vertex don't render it
     If .Features.NbVertex <= 0 Then Exit Sub
     
    If .Modified = False Then
     ' If there is a death time for this mesh, just kill it when this time is reached
     If .DeathTime > 0 Then
       .DeathTime = .DeathTime - SystemFrame
       If .DeathTime <= 0 Then
        .Enabled = False
        Exit Sub
       End If
     End If
    End If
      
      'Check the Sphere Visibility
      Device.SetTransform D3DTS_WORLD, .Matrix
      If .Features.Settings.SphereComputed = True And .Changed = False Then
         If CheckSphere(.Features.Settings.SphereClip, .Features.Settings.SphereRadius) = 0 Then
           .Features.Settings.Visible = 0
           Exit Sub
         End If
      End If
      .Features.Settings.Visible = 1
      
      'Check if we render a part of the mesh
      If RenderPart = False Then
        SGroup = 1
        EGroup = .Features.NumGroup
      Else
        SGroup = StartGroup
        EGroup = EndGroup
        If EndGroup = 0 Then EGroup = .Features.NumGroup
      End If
      
     'Compute the matrix for the mesh (only if the mesh has changed)
     If .Changed = True Or Mesh(MI).Features.Settings.Billboard = True Or .Parent <> 0 Then
       D3DXMatrixScaling ScaleMatrix, .Features.Settings.Scaling.x, .Features.Settings.Scaling.y, .Features.Settings.Scaling.z
       D3DXMatrixTranslation TranslationMatrix, .Features.Settings.Position.x, .Features.Settings.Position.y, .Features.Settings.Position.z
       If Mesh(MI).Features.Settings.Billboard = False Then
         ScaleMatrix = MRotate(ScaleMatrix, .Features.Settings.Rotation.x, .Features.Settings.Rotation.y, .Features.Settings.Rotation.z)
       Else
         ScaleMatrix = MRotate(ScaleMatrix, 0, GetCameraAngle, 0)
       End If
       D3DXMatrixMultiply .Matrix, ScaleMatrix, TranslationMatrix
       If .Parent <> 0 Then
        D3DXMatrixMultiply .Matrix, Mesh(.Parent).Matrix, .Matrix
       End If
       Device.SetTransform D3DTS_WORLD, .Matrix
      .Changed = False
     End If
     
    'Handle Shaders !
    With Device
     
     If Mesh(MI).Features.TypeVertex <> TV_LVERTEX2 Then
       Device.SetRenderState D3DRS_LIGHTING, 1
     Else
       Device.SetRenderState D3DRS_LIGHTING, 0
     End If
       
       
     .SetRenderState D3DRS_ALPHABLENDENABLE, 1
     .SetRenderState D3DRS_ALPHATESTENABLE, 1
    
    If Mesh(MI).Features.Settings.SphereMapping = True Then
       .SetTransform D3DTS_TEXTURE0, SphereMappingMatrix
       .SetRenderState D3DRS_CULLMODE, D3DCULL_NONE
       .SetTextureStageState 0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT2
       .SetTextureStageState 0, D3DTSS_TEXCOORDINDEX, Mesh(MI).Features.Settings.SphereMappingMode
       .SetRenderState D3DRS_NORMALIZENORMALS, 1
    
    Else
       .SetTransform D3DTS_TEXTURE0, WorldMatrix
       .SetTextureStageState 0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE
       .SetTextureStageState 0, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_PASSTHRU
    
    End If
   End With
'     Debug.Print .Features.TypeVertex
     For i = SGroup To EGroup
        If .Features.Groups(i).ShaderI <> 0 Then
         With Shaders(.Features.Groups(i).ShaderI)
          For a = 1 To .NbLayer
            'Simple Map
            
            Select Case .Layers(a).MapType
              Case 0  'simple map
                tex = .Layers(a).tex(1)
              Case 1
                tex = Mesh(MI).Features.Groups(i).LightmapI
              Case 2 'animate
                 If Mesh(MI).Modified = False Then
                 Mesh(MI).Features.Groups(i).ShaderKey = Mesh(MI).Features.Groups(i).ShaderKey + .Layers(a).FPS * SystemFrame / 1000
restart:
                 If Int(Mesh(MI).Features.Groups(i).ShaderKey) > .Layers(a).NbTex Then
                  Mesh(MI).Features.Groups(i).ShaderKey = Mesh(MI).Features.Groups(i).ShaderKey - .Layers(a).NbTex
                  GoTo restart
                 End If
               End If
                tex = .Layers(a).tex(Int(Mesh(MI).Features.Groups(i).ShaderKey))
            End Select
            If .Layers(a).SrcBlend = -1 Then
              Device.SetRenderState D3DRS_SRCBLEND, D3DBLEND_SRCALPHA
              Device.SetRenderState D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA
            Else
              Device.SetRenderState D3DRS_SRCBLEND, .Layers(a).SrcBlend
              Device.SetRenderState D3DRS_DESTBLEND, .Layers(a).DestBlend
            End If
            Mat = .Layers(a).Material
            RenderGroup MI, i, tex, Mat
          Next a
         End With
        Else
          If .Features.Groups(i).AlphaBlending = 0 Or .Features.Groups(i).AlphaBlending = TV_BLEND_ALPHA Then
            Device.SetRenderState D3DRS_SRCBLEND, D3DBLEND_SRCALPHA
            Device.SetRenderState D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA
          Else
           Select Case .Features.Groups(i).AlphaBlending
             Case TV_BLEND_COLOR
                Device.SetRenderState D3DRS_SRCBLEND, D3DBLEND_SRCCOLOR
                Device.SetRenderState D3DRS_DESTBLEND, D3DBLEND_ONE
             Case TV_BLEND_ADD
                Device.SetRenderState D3DRS_SRCBLEND, D3DBLEND_ONE
                Device.SetRenderState D3DRS_DESTBLEND, D3DBLEND_ONE
           End Select
          End If
          RenderGroup MI, i, .Features.Groups(i).TextureI, .Features.Groups(i).MaterialI
        End If
     Next i
    .Modified = True
    End With
 
   
   With Mesh(MI)
    If .NumChild > 0 Then
      For i = 1 To .NumChild
        RenderMesh .Childs(i), False
      Next i
    End If
   End With
   
End Sub

Public Function GetCameraAngle() As Single
 Dim V As D3DVECTOR
 V = VSubtract(EndPos, StartPos)
 
 If V.z > 0 Then
  GetCameraAngle = Atn(V.x / V.z)
 ElseIf V.z = 0 Then
  GetCameraAngle = 0
 Else
  GetCameraAngle = Atn(V.x / V.z) + 3.141592
 End If
  
  GetCameraAngle = GetCameraAngle * 180 / 3.141592

End Function


Public Function GetCameraAngle2(VC As D3DVECTOR) As Single
 Dim V As D3DVECTOR
 V = VSubtract(VC, StartPos)
 
 If V.z > 0 Then
  GetCameraAngle2 = Atn(V.x / V.z)
 ElseIf V.z = 0 Then
  GetCameraAngle2 = 0
 Else
  GetCameraAngle2 = Atn(V.x / V.z) + 3.141592
 End If
  
  GetCameraAngle2 = GetCameraAngle2 * 180 / 3.141592

End Function

Sub SaveMeshToTVO(Filename As String, MI As Integer, Optional SaveTexDesc As Boolean = True)
   Open Filename For Binary As 2
   
      Dim version As Long
      Dim MeshVersion As Long
      Dim C() As ChunkMesh
      Dim t() As Integer, NTex As Integer
      Dim TexTable() As TVOTextureTable
      Dim i As Long, z As Long
      Dim o As Long
      
      ReDim C(0)
      ReDim t(0)
      
      
      Dim TableOffset As Long, TableLength As Long
      version = 8
      
      Put #2, , "TVOF"
      Put #2, , TableOffset  'this stuff will be filled at the end
      AddChunk CHUNK_VERSION, C
      Put #2, , version
      AddChunk CHUNK_MESH, C
      AddChunk CHUNK_MESHNAME, C
      Put #2, , Mesh(MI).MeshName
      AddChunk CHUNK_FEATURES, C
      Put #2, , Mesh(MI).Features.Settings.Position
      Put #2, , Mesh(MI).Features.Settings.Rotation
      Put #2, , Mesh(MI).Features.Settings.Scaling
      Put #2, , Mesh(MI).Features.Settings.SphereClip
      Put #2, , Mesh(MI).Features.Settings.SphereRadius
      
      AddChunk CHUNK_NBVERTEX, C
      Put #2, , Mesh(MI).Features.NbVertex
      If Mesh(MI).Features.TypeVertex = TV_COLOR_VERTEX Then
        AddChunk CHUNK_LVERTEX, C
        Put #2, , Mesh(MI).Features.LVertex()
      Else
        AddChunk CHUNK_VERTEX, C
        Put #2, , Mesh(MI).Features.Vertex()
      End If
      
      'Write some mesh settings
      If Mesh(MI).Features.Settings.Billboard = True Then
        AddChunk CHUNK_BILLBOARD, C
      End If
      
      If Mesh(MI).Features.Settings.SPREnabled = True Then
        AddChunk CHUNK_SPRITE, C
      End If
      
      'Write face number
      AddChunk CHUNK_NBFACE, C
      Put #2, , Mesh(MI).Features.NumGroup
      
      'Write face descs
      For i = 1 To Mesh(MI).Features.NumGroup
        With Mesh(MI).Features.Groups(i)
        AddChunk CHUNK_FACE, C
        AddChunk CHUNK_SPHEREMAPPING, C
        Put #2, , Mesh(MI).Features.Groups(i).ShpereMapping
        AddChunk CHUNK_ALPHABLENDING, C
        Put #2, , Mesh(MI).Features.Groups(i).AlphaBlending
      
      If SaveTexDesc = True Then
        If .LightmapI > 0 Then
          AddChunk CHUNK_FACELIGHTMAP, C
          Put #2, , .LightmapI
        End If
        
        If .MaterialI > 0 Then
          AddChunk CHUNK_FACEMATERIAL, C
          Put #2, , .MaterialI
        End If
        
        If .TextureI > 0 Then
          'Check if this texture is already in texture table
          For o = 1 To NTex
            If .TextureI = t(o) Then
             Exit For
            End If
          Next o
          If o = NTex + 1 Then
            NTex = NTex + 1
            ReDim Preserve t(NTex)
          End If
          
          'Store the used texture index
          t(o) = .TextureI
          
          AddChunk CHUNK_FACETEXTURE, C
          Put #2, , .TextureI
        End If
      End If
        
        AddChunk CHUNK_FACEVERTEXNUMBER, C
        Put #2, , .IndicesCount
        
        AddChunk CHUNK_FACEVERTEXSTART, C
        Put #2, , .VertexOffset
        
        AddChunk CHUNK_ENDFACE, C
        End With
      Next i
      
      AddChunk CHUNK_ENDMESH, C
      
      'TODO :Write the texture table
      If SaveTexDesc = True Then
        AddChunk CHUNK_NBTEXTURE, C
        Put #2, , NTex
        AddChunk CHUNK_TEXTURE, C
        'Write the texture table
        ReDim TexTable(NTex)
        
        For o = 1 To NTex
         TexTable(o).Filename = Textures(t(o)).TextureFileName
         TexTable(o).TextureName = Textures(t(o)).TextureName
         TexTable(o).width = Textures(t(o)).TextureWidth
         TexTable(o).height = Textures(t(o)).TextureHeight
        Next o
        
        'Put it in the file
        Put #2, , TexTable
      End If
    
      'Write the chunk table
      TableOffset = Seek(2)
      Put #2, , C
      Put #2, 5, TableOffset
   Close #2
End Sub
Function Trime(Str As String) As String
  Dim ins As Long
  Do
    If ins <> 0 Then
       If ins = 1 Then
         Str = mid$(Str, ins + 1)
       Else
         Str = Left$(Str, ins - 1) + mid$(Str, ins + 1)
       End If
    End If
    Str = Trim$(Str)
    ins = InStr(Str, vbTab)
  Loop While ins <> 0
  Trime = Str
End Function

Function VolumeCheck(V As D3DVECTOR, M As D3DMATRIX) As Byte
    Dim v4 As D3DVECTOR, Mask As Byte, i As Long
    Dim Dist As Single
    'DXCopyMemory V4, v, 12
    Mask = 0
   ' V4.w = 1
   ' Debug.Print V4.x, V4.y, V4.z, V4.w
   ' D3DXVec4Transform V4, V4, M
    v4 = MTransform(M, V)
   '' Debug.Print V4.x, V4.y, V4.z, V4.w'
   For i = 0 To 5
        Dist = D3DXPlaneDotCoord(FrustumPlanes(i), v4)
        If Dist < 0 Then
           Mask = Mask Or (2 ^ i)
        End If
   Next i
  ' CheckSphere = 1
 
 '   If V4.x > V4.w Then Mask = Mask Or TV_PLANE_RIGHT
 '   If V4.x < -V4.w Then Mask = Mask Or TV_PLANE_LEFT
 '   If V4.y > V4.w Then Mask = Mask Or TV_PLANE_TOP
 '   If V4.y > -V4.w Then Mask = Mask Or TV_PLANE_BOTTOM
 '   If V4.z > V4.w Then Mask = Mask Or TV_PLANE_FAR
 '   If V4.z > -V4.w Then Mask = Mask Or TV_PLANE_NEAR
    VolumeCheck = Mask
    
    
End Function


